import sys
import wx
import wx.lib.mixins.inspection

from wxutils import SetTip

FRAME_BG = (253, 253, 250) ## light grey
TITLE_BG = (253, 253, 250) ## light grey
FGCOL    = ( 20,  20, 120) ## blue
BGCOL    = (253, 253, 250) ## light grey
BGSEL    = (250, 250, 200) ## yellow
FGSEL    = (200,   0,   0) ## red

class PeriodicTablePanel(wx.Panel):
    """periodic table of the elements"""
    elems = {'H':  ( 0,  0), 'He': ( 0, 17), 'Li': ( 1,  0), 'Be': ( 1,  1),
             'B':  ( 1, 12), 'C':  ( 1, 13), 'N':  ( 1, 14), 'O':  ( 1, 15),
             'F':  ( 1, 16), 'Ne': ( 1, 17), 'Na': ( 2,  0), 'Mg': ( 2,  1),
             'Al': ( 2, 12), 'Si': ( 2, 13), 'P':  ( 2, 14), 'S':  ( 2, 15),
             'Cl': ( 2, 16), 'Ar': ( 2, 17), 'K':  ( 3,  0), 'Ca': ( 3,  1),
             'Sc': ( 3,  2), 'Ti': ( 3,  3), 'V':  ( 3,  4), 'Cr': ( 3,  5),
             'Mn': ( 3,  6), 'Fe': ( 3,  7), 'Co': ( 3,  8), 'Ni': ( 3,  9),
             'Cu': ( 3, 10), 'Zn': ( 3, 11), 'Ga': ( 3, 12), 'Ge': ( 3, 13),
             'As': ( 3, 14), 'Se': ( 3, 15), 'Br': ( 3, 16), 'Kr': ( 3, 17),
             'Rb': ( 4,  0), 'Sr': ( 4,  1), 'Y':  ( 4,  2), 'Zr': ( 4,  3),
             'Nb': ( 4,  4), 'Mo': ( 4,  5), 'Tc': ( 4,  6), 'Ru': ( 4,  7),
             'Rh': ( 4,  8), 'Pd': ( 4,  9), 'Ag': ( 4, 10), 'Cd': ( 4, 11),
             'In': ( 4, 12), 'Sn': ( 4, 13), 'Sb': ( 4, 14), 'Te': ( 4, 15),
             'I':  ( 4, 16), 'Xe': ( 4, 17), 'Cs': ( 5,  0), 'Ba': ( 5,  1),
             'La': ( 5,  2), 'Ce': ( 7,  3), 'Pr': ( 7,  4), 'Nd': ( 7,  5),
             'Pm': ( 7,  6), 'Sm': ( 7,  7), 'Eu': ( 7,  8), 'Gd': ( 7,  9),
             'Tb': ( 7, 10), 'Dy': ( 7, 11), 'Ho': ( 7, 12), 'Er': ( 7, 13),
             'Tm': ( 7, 14), 'Yb': ( 7, 15), 'Lu': ( 7, 16), 'Hf': ( 5,  3),
             'Ta': ( 5,  4), 'W':  ( 5,  5), 'Re': ( 5,  6), 'Os': ( 5,  7),
             'Ir': ( 5,  8), 'Pt': ( 5,  9), 'Au': ( 5, 10), 'Hg': ( 5, 11),
             'Tl': ( 5, 12), 'Pb': ( 5, 13), 'Bi': ( 5, 14), 'Po': ( 5, 15),
             'At': ( 5, 16), 'Rn': ( 5, 17), 'Fr': ( 6,  0), 'Ra': ( 6,  1),
             'Ac': ( 6,  2), 'Th': ( 8,  3), 'Pa': ( 8,  4), 'U':  ( 8,  5),
             'Np': ( 8,  6), 'Pu': ( 8,  7), 'Am': ( 8,  8), 'Cm': ( 8,  9),
             'Bk': ( 8, 10), 'Cf': ( 8, 11), 'Es': ( 8, 12), 'Fm': ( 8, 13),
             'Md': ( 8, 14), 'No': ( 8, 15), 'Lr': ( 8, 16)}

    syms = ['H', 'He', 'Li', 'Be', 'B', 'C', 'N', 'O', 'F', 'Ne', 'Na', 'Mg',
            'Al', 'Si', 'P', 'S', 'Cl', 'Ar', 'K', 'Ca', 'Sc', 'Ti', 'V',
            'Cr', 'Mn', 'Fe', 'Co', 'Ni', 'Cu', 'Zn', 'Ga', 'Ge', 'As', 'Se',
            'Br', 'Kr', 'Rb', 'Sr', 'Y', 'Zr', 'Nb', 'Mo', 'Tc', 'Ru', 'Rh',
            'Pd', 'Ag', 'Cd', 'In', 'Sn', 'Sb', 'Te', 'I', 'Xe', 'Cs', 'Ba',
            'La', 'Ce', 'Pr', 'Nd', 'Pm', 'Sm', 'Eu', 'Gd', 'Tb', 'Dy', 'Ho',
            'Er', 'Tm', 'Yb', 'Lu', 'Hf', 'Ta', 'W', 'Re', 'Os', 'Ir', 'Pt',
            'Au', 'Hg', 'Tl', 'Pb', 'Bi', 'Po', 'At', 'Rn', 'Fr', 'Ra', 'Ac',
            'Th', 'Pa', 'U', 'Np', 'Pu', 'Am', 'Cm', 'Bk', 'Cf', 'Es', 'Fm',
            'Md', 'No', 'Lr']


    names = ['hydrogen', 'helium', 'lithium', 'beryllium', 'boron',
             'carbon', 'nitrogen', 'oxygen', 'fluorine', 'neon', 'sodium',
             'magnesium', 'aluminum', 'silicon', 'phosphorus', 'sulfur',
             'chlorine', 'argon', 'potassium', 'calcium', 'scandium',
             'titanium', 'vanadium', 'chromium', 'manganese', 'iron',
             'cobalt', 'nickel', 'copper', 'zinc', 'gallium', 'germanium',
             'arsenic', 'selenium', 'bromine', 'krypton', 'rubidium',
             'strontium', 'yttrium', 'zirconium', 'niobium', 'molybdenum',
             'technetium', 'ruthenium', 'rhodium', 'palladium', 'silver',
             'cadmium', 'indium', 'tin', 'antimony', 'tellurium', 'iodine',
             'xenon', 'cesium', 'barium', 'lanthanum', 'cerium',
             'praseodymium', 'neodymium', 'promethium', 'samarium',
             'europium', 'gadolinium', 'terbium', 'dysprosium', 'holmium',
             'erbium', 'thulium', 'ytterbium', 'lutetium', 'hafnium',
             'tantalum', 'tungsten', 'rhenium', 'osmium', 'iridium',
             'platinum', 'gold', 'mercury', 'thallium', 'lead', 'bismuth',
             'polonium', 'astatine', 'radon', 'francium', 'radium',
             'actinium', 'thorium', 'protactinium', 'uranium', 'neptunium',
             'plutonium', 'americium', 'curium', 'berkelium', 'californium',
             'einsteinium', 'fermium', 'mendelevium', 'nobelium',
             'lawrencium']


    def __init__(self, parent, title='Select Element', multi_select=False,
                 onselect=None, tooltip_msg=None, size=(-1, -1), fontsize=10,
                 fgcol=None, bgcol=None, fgsel=None, bgsel=None, **kws):
        wx.Panel.__init__(self, parent, -1, size=size, name='PeriodicTable', **kws)
        self.parent = parent
        self.onselect = onselect
        self.tooltip_msg = tooltip_msg
        self.wids = {}
        self.ctrls = {}
        self.SetBackgroundColour(FRAME_BG)
        self.selected = []
        if fgcol is None: fgcol = FGCOL
        if bgcol is None: bgcol = BGCOL
        if fgsel is None: fgsel = FGSEL
        if bgsel is None: bgsel = BGSEL
        self.fgcol = fgcol
        self.bgcol = bgcol
        self.fgsel = fgsel
        self.bgsel = bgsel

        self.current = None
        self.multi_select = multi_select
        if sys.platform.lower().startswith('win'):
            fonstize = fontsize - 1
        subfontsize = fontsize
        if sys.platform.lower().startswith('lin'):
            subfontsize -= 1
        self.titlefont = wx.Font(fontsize, wx.DEFAULT, wx.NORMAL, wx.BOLD)
        self.elemfont = wx.Font(fontsize,  wx.SWISS, wx.NORMAL, wx.BOLD)
        self.subtitlefont = wx.Font(subfontsize, wx.DEFAULT, wx.NORMAL, wx.BOLD)

        self.BuildPanel()

    def onKey(self, event=None, name=None):
        """support browsing through elements with arrow keys"""
        if self.current is None:
            return
        selname = self.current.GetLabel()
        if selname  in self.elems:
            coords = self.elems[selname]
            if name is None and event is not None:
                thiskey = event.GetKeyCode()
            if name == 'left':
                thiskey = wx.WXK_LEFT
            elif name == 'right':
                thiskey = wx.WXK_RIGHT
            elif name == 'up':
                thiskey = wx.WXK_UP
            elif name == 'down':
                thiskey = wx.WXK_DOWN
            newcoords = None
            if thiskey == wx.WXK_UP:
                newcoords = (coords[0]-1, coords[1])
            elif thiskey == wx.WXK_DOWN:
                newcoords = (coords[0]+1, coords[1])
            elif thiskey in (wx.WXK_LEFT, wx.WXK_RIGHT):
                newcoords = None
            # try to support jumping to/from lanthanide,
            # and wrapping around elements
            if newcoords not in self.elems.values():
                if thiskey == wx.WXK_DOWN:
                    newcoords = (coords[0]+2, coords[1])
                elif thiskey == wx.WXK_UP:
                    newcoords = (coords[0]-2, coords[1])
                elif thiskey in (wx.WXK_LEFT, wx.WXK_RIGHT):
                    try:
                        znum = self.syms.index(selname)
                    except:
                        return
                    if thiskey == wx.WXK_LEFT and znum > 0:
                        newcoords = self.elems[self.syms[znum-1]]
                    elif thiskey == wx.WXK_RIGHT and znum < len(self.syms)-1:
                        newcoords = self.elems[self.syms[znum+1]]

            if newcoords in self.elems.values():
                newlabel = None
                for xlabel, xcoords in self.elems.items():
                    if newcoords == xcoords:
                        newlabel = xlabel
                if newlabel is not None:
                    self.onclick(label=newlabel)
        event.Skip()

    def on_clear_all(self, event=None):
        for wid in list(self.ctrls.values()) + list(self.wids.values()):
            wid.SetBackgroundColour(self.bgcol)
            wid.SetForegroundColour(self.fgcol)
        self.selected = []
        self.current = None

    def onclick(self, event=None, label=None):
        if event is None and label is None:
            return
        textwid = None
        if label is None and event.Id in self.wids:
            textwid = self.wids[event.Id]
            label = textwid.GetLabel()
        if label is None:
            return
        if textwid is None and label is not None:
            textwid = self.ctrls[label]

        if self.multi_select:
            if label in self.selected: # already selected
                textwid.SetBackgroundColour(self.bgcol)
                textwid.SetForegroundColour(self.fgcol)
                self.selected.remove(label)
            else:
                textwid.SetBackgroundColour(self.bgsel)
                textwid.SetForegroundColour(self.fgsel)
                self.selected.append(label)

        else:
            textwid.SetBackgroundColour(self.bgsel)
            textwid.SetForegroundColour(self.fgsel)
            if self.current is not None and self.current != textwid:
                self.current.SetBackgroundColour(self.bgcol)
                self.current.SetForegroundColour(self.fgcol)
            self.current = textwid
            self.selected = [textwid]

        znum = self.syms.index(label)
        name = self.names[znum]

        self.tsym.SetLabel(label)
        self.title.SetLabel(name)
        self.tznum.SetLabel("{:d}".format(znum+1))
        if self.onselect is not None:
            self.onselect(elem=label, event=event)
        self.Refresh()

    def BuildPanel(self):
        sizer = wx.GridBagSizer()
        sizer.SetHGap(0)
        sizer.SetVGap(1)

        for name, coords in self.elems.items():
            tw = wx.StaticText(self, -1, label=name)
            tw.SetFont(self.elemfont)
            tw.SetForegroundColour(self.fgcol)
            tw.SetBackgroundColour(self.bgcol)
            tw.SetMinSize((20, 18))
            tw.Bind(wx.EVT_LEFT_DOWN, self.onclick)
            if self.tooltip_msg is not None:
                SetTip(tw, self.tooltip_msg)
            self.wids[tw.Id] = tw
            self.ctrls[name] = tw
            sizer.Add(tw, coords, (1, 1), wx.ALIGN_LEFT, 0)
        self.title = wx.StaticText(self, -1, label=' Select Element ')
        self.tsym  = wx.StaticText(self, -1, label='__')
        self.tznum = wx.StaticText(self, -1, label='__')

        for a in (self.title, self.tsym, self.tznum):
            a.SetFont(self.titlefont)
            a.SetBackgroundColour(TITLE_BG)

        sizer.Add(self.title, (0, 4), (1, 8), wx.ALIGN_CENTER, 5)
        sizer.Add(self.tsym,  (0, 2), (1, 2), wx.ALIGN_LEFT, 5)
        sizer.Add(self.tznum, (0, 12), (1, 3), wx.ALIGN_LEFT, 5)

        self.subtitle = [None, None]
        self.subtitle[0] = wx.StaticText(self, -1, label='         ')
        self.subtitle[0].SetFont(self.subtitlefont)
        self.subtitle[1] = wx.StaticText(self, -1, label='         ')
        self.subtitle[1].SetFont(self.subtitlefont)

        sizer.Add(self.subtitle[0], (1, 2), (1, 9), wx.ALIGN_LEFT)
        sizer.Add(self.subtitle[1], (2, 2), (1, 9), wx.ALIGN_LEFT)

        sizer.SetEmptyCellSize((2, 2))
        self.Bind(wx.EVT_KEY_UP, self.onKey)
        #self.Bind(wx.EVT_SET_FOCUS, self.onGetFocus)
        #self.Bind(wx.EVT_KILL_FOCUS, self.onLoseFocus)
        self.SetSizer(sizer)
        ix, iy = self.GetBestSize()
        self.SetSize((ix+2, iy+2))
        sizer.Fit(self)

    def onLoseFocus(self, event=None):
        event.Skip()

    def onGetFocus(self, event=None):
        event.Skip()

    def set_subtitle(self, label, index=0):
        if index not in (0, 1): index = 0
        self.subtitle[index].SetLabel(label)

class PTableFrame(wx.Frame):
    def __init__(self, size=(-1, -1)):
        wx.Frame.__init__(self, parent=None, size=size)
        ptab  = PeriodicTablePanel(self, title='Periodic Table',
                                   tooltip_msg='Select Element',
                                   onselect=self.onElement)
        sx, sy = ptab.GetBestSize()
        sizer = wx.BoxSizer(wx.VERTICAL)
        sizer.Add(ptab, 1, wx.EXPAND|wx.ALL, 5)
        self.SetSizer(sizer)
        sizer.Fit(self)
        self.SetMinSize((sx+10, sy+10))
        self.Raise()

    def onElement(self, elem=None, event=None):
        print( 'Element Selected:  ', elem)

class PTableApp(wx.App, wx.lib.mixins.inspection.InspectionMixin):
    def __init__(self, **kws):
        wx.App.__init__(self)

    def OnInit(self):
        self.Init()
        frame = PTableFrame() #
        frame.Show()
        self.SetTopWindow(frame)
        self.ShowInspectionTool()
        return True

if __name__ == "__main__":
    PTableApp().MainLoop()



#!/usr/bin/env python
'''
Diffraction functions require for fitting and analyzing data.

mkak 2017.02.06 (originally written spring 2016)
'''

##########################################################################
# IMPORT PYTHON PACKAGES

import math
import numpy as np

from ..utils.physical_constants import PLANCK_HC

import re
import os
import cmath

HAS_CifFile = False
try:
    import CifFile
    HAS_CifFile = True
except ImportError:
    pass

##########################################################################
# FUNCTIONS

def d_from_q(q):
    '''
    Converts q axis into d (returned units inverse of provided units)
    d = 2*PI/q
    '''
    return (2.*math.pi)/q

def d_from_twth(twth,wavelength,ang_units='degrees'):
    '''
    Converts 2th axis into d (returned units same as wavelength units)
    d = lambda/[2*sin(2th/2)]

    ang_unit : default in degrees; will convert from 'rad' if given
    '''
    if not ang_units.startswith('rad'):
        twth = np.radians(twth)
    return wavelength/(2*np.sin(twth/2.))


def twth_from_d(d,wavelength,ang_units='degrees'):
    '''
    Converts d axis into 2th (d and wavelength must be in same units)
    2th = 2*sin^-1(lambda/[2*d])

    ang_unit : default in degrees; will convert to 'rad' if given
    '''
    twth = 2.*np.arcsin(wavelength/(2.*d))
    if ang_units.startswith('rad'):
        return twth
    else:
        return np.degrees(twth)


def twth_from_q(q,wavelength,ang_units='degrees'):
    '''
    Converts q axis into 2th (q and wavelength will have inverse units)
    2th = 2*sin^-1(lambda/[2*d])

    ang_unit : default in degrees; will convert to 'rad' if given
    '''
    twth = 2.*np.arcsin((q*wavelength)/(4.*math.pi))
    if ang_units.startswith('rad'):
        return twth
    else:
        return np.degrees(twth)


def q_from_d(d):
    '''
    Converts d axis into q (returned units inverse of provided units)
    q = 2*PI/d
    '''
    return (2.*math.pi)/d


def q_from_twth(twth,wavelength,ang_units='degrees'):
    '''
    Converts 2th axis into q (q returned in inverse units of wavelength)
    q = [(4*PI)/lamda]*sin(2th/2)

    ang_unit : default in degrees; will convert from 'rad' if given
    '''
    if not ang_units.startswith('rad'):
        twth = np.radians(twth)
    return ((4.*math.pi)/wavelength)*np.sin(twth/2.)

def qv_from_hkl(hklall,a,b,c,alp,bet,gam):

    qv = np.zeros(np.shape(hklall))
    uvol = unit_cell_volume(a,b,c,alp,bet,gam)

    alp,bet,gam = np.radians(alp),np.radians(bet),np.radians(gam)
    q0 = [(b*c*np.sin(alp))/uvol,(c*a*np.sin(bet))/uvol,(a*b*np.sin(gam))/uvol]

    for i,hkl in enumerate(hklall):
        qv[i] = [2*math.pi*hkl[0]*q0[0],
                 2*math.pi*hkl[1]*q0[1],
                 2*math.pi*hkl[2]*q0[2]]
    return qv

def d_from_hkl(hklall,a,b,c,alp,bet,gam):

    d = np.zeros(len(hklall))
    alp,bet,gam = np.radians(alp),np.radians(bet),np.radians(gam)
    for i,hkl in enumerate(hklall):
        h,k,l = hkl
        x = 1-np.cos(alp)**2 - np.cos(bet)**2 - np.cos(gam)**2 \
                + 2*np.cos(alp)*np.cos(bet)*np.cos(gam)
        y =   (h*np.sin(alp)/a)**2 + 2*k*l*(np.cos(bet)*np.cos(gam)-np.cos(alp))/(b*c) \
            + (k*np.sin(bet)/b)**2 + 2*l*h*(np.cos(gam)*np.cos(alp)-np.cos(bet))/(c*a) \
            + (l*np.sin(gam)/c)**2 + 2*h*k*(np.cos(alp)*np.cos(bet)-np.cos(gam))/(a*b)
        d[i] = np.sqrt(x/y)

    return d

def unit_cell_volume(a,b,c,alp,bet,gam):

    alp,bet,gam = np.radians(alp),np.radians(bet),np.radians(gam)
    return a*b*c*(1-np.cos(alp)**2-np.cos(bet)**2-np.cos(gam)**2+2*np.cos(alp)*np.cos(bet)*np.cos(gam))**0.5


def E_from_lambda(wavelength,E_units='keV',lambda_units='A'):
    '''
    Converts lambda into energy
    E = hf ; E = hc/lambda

    E_units      : default keV; can convert to 'eV' if given
    lambda_units : default 'A'; can convert from 'm' or 'nm' if given
    '''
    if lambda_units == 'm':
        wavelength = wavelength*1e10
    elif lambda_units == 'nm':
        wavelength = wavelength*1e1
    if E_units.lower() == 'kev':
        return PLANCK_HC/wavelength*1e-3 # keV
    else:
        return (PLANCK_HC/wavelength)    # eV


def lambda_from_E(E, E_units='keV', lambda_units='A'):
    '''
    Converts lambda into energy
    E = hf ; E = hc/lambda

    E_units      : default keV; can convert from 'eV' if given
    lambda_units : default 'A'; can convert to 'm' or 'nm' if given
    '''
    if E_units.lower() != 'kev':
        E = E*1e-3 # keV
    if lambda_units == 'm':
        return (PLANCK_HC/E)*1e-13
    elif lambda_units == 'nm':
        return (PLANCK_HC/E)*1e-4
    else:
        return PLANCK_HC/E*1e-3 # A

def generate_hkl(hmax=10,kmax=10,lmax=10,positive_only=True):
    if positive_only is True:
        hklall = np.mgrid[0:hmax+1, 0:kmax+1, 0:lmax+1].reshape(3, -1).T
    else:
        hklall = np.mgrid[-hmax:hmax+1, -kmax:kmax+1, -lmax:lmax+1].reshape(3, -1).T
    return np.array([hkl for hkl in hklall if hkl[0]**2 + hkl[1]**2 + hkl[2]**2 > 0])



#!/usr/bin/env python
'''
popup frame for 2D XRD image control

'''
import numpy as np
import matplotlib.cm as colormap
import wx

###################################

class ImageToolboxFrame(wx.Frame):

    def __init__(self,imageframe,image): #,mask=False,bkgd=False):
        '''
        Frame for visual toolbox
        '''
        label = 'Image Toolbox'
        wx.Frame.__init__(self, None, -1,title=label, size=(330, 300))

        #self.SetMinSize((700,500))

        ## Set inputs
        self.plot2Dframe = imageframe
        #self.raw_img = image
        self.plt_img = image

        ## Set defaults
        self.bkgd_scale = 0

        self.color = 'bone'
        self.flip = 'vertical'

        self.Init()
        self.Centre()
        self.Show(True)

    def Init(self):

        self.panel = wx.Panel(self)
        vbox = wx.BoxSizer(wx.VERTICAL)

        ###########################
        ## Color
        hbox_clr = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_clr = wx.StaticText(self.panel, label='COLOR')
        colors = []
        for key in colormap.datad:
            if not key.endswith('_r'):
                colors.append(key)
        self.ch_clr = wx.Choice(self.panel,choices=colors)

        self.ch_clr.Bind(wx.EVT_CHOICE,self.onColor)

        hbox_clr.Add(self.txt_clr, flag=wx.RIGHT, border=8)
        hbox_clr.Add(self.ch_clr, flag=wx.EXPAND, border=8)
        vbox.Add(hbox_clr, flag=wx.ALL, border=10)

        ###########################
        ## Contrast
        vbox_ct = wx.BoxSizer(wx.VERTICAL)

        hbox_ct1 = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_ct1 = wx.StaticText(self.panel, label='CONTRAST')
        self.txt_ct2 = wx.StaticText(self.panel, label='')

        hbox_ct1.Add(self.txt_ct1, flag=wx.EXPAND|wx.RIGHT, border=8)
        hbox_ct1.Add(self.txt_ct2, flag=wx.ALIGN_RIGHT, border=8)
        vbox_ct.Add(hbox_ct1, flag=wx.BOTTOM, border=8)

        hbox_ct2 = wx.BoxSizer(wx.HORIZONTAL)
        self.ttl_min = wx.StaticText(self.panel, label='min')
        self.sldr_min = wx.Slider(self.panel, style=wx.SL_LABELS, maxValue=5e6)
        self.entr_min = wx.TextCtrl(self.panel,wx.TE_PROCESS_ENTER)

        self.sldr_min.Bind(wx.EVT_SLIDER,self.onSlider)

        hbox_ct2.Add(self.ttl_min, flag=wx.RIGHT, border=8)
        hbox_ct2.Add(self.sldr_min, flag=wx.EXPAND, border=8)
        hbox_ct2.Add(self.entr_min, flag=wx.RIGHT, border=8)
        vbox_ct.Add(hbox_ct2, flag=wx.BOTTOM, border=8)

        hbox_ct3 = wx.BoxSizer(wx.HORIZONTAL)
        self.ttl_max = wx.StaticText(self.panel, label='max')
        self.sldr_max = wx.Slider(self.panel, style=wx.SL_LABELS, maxValue=5e6)
        self.entr_max = wx.TextCtrl(self.panel,wx.TE_PROCESS_ENTER)

        self.sldr_max.Bind(wx.EVT_SLIDER,self.onSlider)

        hbox_ct3.Add(self.ttl_max, flag=wx.RIGHT, border=8)
        hbox_ct3.Add(self.sldr_max, flag=wx.EXPAND, border=8)
        hbox_ct3.Add(self.entr_max, flag=wx.RIGHT, border=8)
        vbox_ct.Add(hbox_ct3, flag=wx.BOTTOM, border=8)

        hbox_ct4 = wx.BoxSizer(wx.HORIZONTAL)
        self.btn_ct1 = wx.Button(self.panel,label='reset range')
        self.btn_ct2 = wx.Button(self.panel,label='set range')

        self.btn_ct1.Bind(wx.EVT_BUTTON,self.autoContrast)
        self.btn_ct2.Bind(wx.EVT_BUTTON,self.onContrastRange)

        hbox_ct4.Add(self.btn_ct1, flag=wx.RIGHT, border=8)
        hbox_ct4.Add(self.btn_ct2, flag=wx.RIGHT, border=8)
        vbox_ct.Add(hbox_ct4, flag=wx.ALIGN_RIGHT|wx.BOTTOM,border=8)
        vbox.Add(vbox_ct, flag=wx.ALL, border=10)



        ###########################
        ## Flip
        hbox_flp = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_flp = wx.StaticText(self.panel, label='IMAGE FLIP')
        flips = ['none','vertical','horizontal','both']
        self.ch_flp = wx.Choice(self.panel,choices=flips)

        self.ch_flp.Bind(wx.EVT_CHOICE,self.onFlip)

        hbox_flp.Add(self.txt_flp, flag=wx.RIGHT, border=8)
        hbox_flp.Add(self.ch_flp, flag=wx.EXPAND, border=8)
        vbox.Add(hbox_flp, flag=wx.ALL, border=10)

        ###########################
        ## Scale
        hbox_scl = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_scl = wx.StaticText(self.panel, label='SCALE')
        scales = ['linear','log']
        self.ch_scl = wx.Choice(self.panel,choices=scales)

        self.ch_scl.Bind(wx.EVT_CHOICE,self.onScale)

        hbox_scl.Add(self.txt_scl, flag=wx.RIGHT, border=8)
        hbox_scl.Add(self.ch_scl, flag=wx.EXPAND, border=8)
        vbox.Add(hbox_scl, flag=wx.ALL, border=10)


        ###########################
        ## Set defaults
        self.ch_clr.SetStringSelection(self.color)
        self.ch_flp.SetStringSelection(self.flip)
        self.setSlider()

        self.panel.SetSizer(vbox)

    def autoContrast(self,event=None):

        self.minINT = int(np.min(self.plt_img))
        self.maxINT = int(np.max(self.plt_img)/15) # /15 scales image to viewable
        if self.maxINT == self.minINT:
            self.minINT = self.minINT
            self.maxINT = self.minINT+100
        try:
            self.sldr_min.SetRange(self.minINT,self.maxINT)
            self.sldr_max.SetRange(self.minINT,self.maxINT)
        except:
            pass
        self.minCURRENT = self.minINT
        self.maxCURRENT = self.maxINT
        if self.maxCURRENT > self.maxINT:
            self.maxCURRENT = self.maxINT
        self.setContrast()

    def onContrastRange(self,event=None):

        newMIN = int(self.entr_min.GetValue())
        newMAX = int(self.entr_max.GetValue())

        self.minCURRENT = newMIN
        self.maxCURRENT = newMAX

        self.sldr_min.SetRange(newMIN,newMAX)
        self.sldr_max.SetRange(newMIN,newMAX)

        self.setContrast()

    def setSlider(self):

        self.minCURRENT = self.plot2Dframe.conf.int_lo[0]
        self.maxCURRENT = self.plot2Dframe.conf.int_hi[0]
#         self.minCURRENT = self.plot2Dframe.conf.int_lo['int']
#         self.maxCURRENT = self.plot2Dframe.conf.int_hi['int']

        self.entr_min.SetLabel(str(self.minCURRENT))
        self.entr_max.SetLabel(str(self.maxCURRENT))

        self.minINT = int(np.min(self.plt_img))
        self.maxINT = int(np.max(self.plt_img)/15) # /15 scales image to viewable
        if self.maxINT == self.minINT:
            self.minINT = self.minINT
            self.maxINT = self.minINT+100
        try:
            self.sldr_min.SetRange(self.minINT,self.maxINT)
            self.sldr_max.SetRange(self.minINT,self.maxINT)
        except:
            pass

        try:
            self.sldr_min.SetValue(self.minCURRENT)
            self.sldr_max.SetValue(self.maxCURRENT)
        except:
            pass

    def onSlider(self,event=None):

        self.minCURRENT = self.sldr_min.GetValue()
        self.maxCURRENT = self.sldr_max.GetValue()

        ## Create safety to keep min. below max.
        ## mkak 2016.10.20

        self.setContrast()

    def setContrast(self):

        self.sldr_min.SetValue(self.minCURRENT)
        self.sldr_max.SetValue(self.maxCURRENT)

        self.plot2Dframe.conf.auto_intensity = False
        self.plot2Dframe.conf.int_lo[0] = self.minCURRENT
        self.plot2Dframe.conf.int_hi[0] = self.maxCURRENT
#         self.plot2Dframe.conf.int_lo['int'] = self.minCURRENT
#         self.plot2Dframe.conf.int_hi['int'] = self.maxCURRENT

        self.plot2Dframe.redraw()

        self.entr_min.SetLabel(str(self.minCURRENT))
        self.entr_max.SetLabel(str(self.maxCURRENT))

    def onFlip(self,event=None):
        '''
        Eventually, should just set self.raw_img or self.fli_img - better than this
        mkak 2016.10.20
        '''

        if self.ch_flp.GetString(self.ch_flp.GetSelection()) != self.flip:
            self.flip = self.ch_flp.GetString(self.ch_flp.GetSelection())

            self.checkFLIPS()

            self.plot2Dframe.redraw()

    def checkFLIPS(self):

        if self.flip == 'vertical': # Vertical
            self.plot2Dframe.conf.flip_ud = True
            self.plot2Dframe.conf.flip_lr = False
        elif self.flip == 'horizontal': # Horizontal
            self.plot2Dframe.conf.flip_ud = False
            self.plot2Dframe.conf.flip_lr = True
        elif self.flip == 'both': # both
            self.plot2Dframe.conf.flip_ud = True
            self.plot2Dframe.conf.flip_lr = True
        else: # None
            self.plot2Dframe.conf.flip_ud = False
            self.plot2Dframe.conf.flip_lr = False

    def onScale(self,event=None):
        if self.ch_scl.GetSelection() == 1: ## log
            self.plot2Dframe.conf.log_scale = True
        else:  ## linear
            self.plot2Dframe.conf.log_scale = False
        self.plot2Dframe.redraw()

    def onColor(self,event=None):
        if self.color != self.ch_clr.GetString(self.ch_clr.GetSelection()):
            self.color = self.ch_clr.GetString(self.ch_clr.GetSelection())
            self.setColor()

    def setColor(self):
        self.plot2Dframe.conf.cmap[0] = getattr(colormap, self.color)
#         self.plot2Dframe.conf.cmap['int'] = getattr(colormap, self.color)
        self.plot2Dframe.display(self.plt_img)
        self.checkFLIPS()
        self.plot2Dframe.redraw()


    def SetContrast(self):

        self.txt_ct2.SetLabel('[ full range: %i, %i ]' %
                  (np.min(self.plt_img),np.max(self.plt_img)))

        self.autoContrast()
        self.checkFLIPS()

        self.plot2Dframe.redraw()

#############################################################
################    IN PROGRESS - START   ###################
#############################################################

# class ImageToolboxPanel(wx.Panel):
#
#     def __init__(self,parent):#,imageframe,image,mask=False,bkgd=False):
#         '''
#         Panel for visual toolbox
#         '''
#         label = 'Image Toolbox'
#         wx.Panel.__init__(self, parent, -1)
#         #wx.Panel.__init__(self, None, -1)
#         #wx.Panel.__init__(self, parent, -1, **kws)
#
#         #self.SetMinSize((700,500))
#
#         self.plot2Dframe = None
#         self.raw_img = None
#         self.mask = False
#         self.bkgd = False
#
#         self.Init()
#
#     def Init(self):
#
#         self.panel = wx.Panel(self)
#         vbox = wx.BoxSizer(wx.VERTICAL)
#
#
# #############################################################
# ################     IN PROGRESS - END    ###################
# #############################################################



from .XRD2Dviewer import XRD2DViewerFrame, XRD2DViewer
from .XRD1Dviewer import XRD1DViewerFrame, XRD1DViewer, Calc1DPopup, plot_sticks
from .ImageControlsFrame import ImageToolboxFrame
from .XRDCalibrationFrame import CalibrationPopup, CalXRD
from .XRDMaskFrame import MaskToolsPopup



from ..wxlib import (gui_utils, ParameterPanel, PeriodicTablePanel,
                     _newplot, _plot, XRFDisplayFrame)

from .gse_dtcorrect import DTViewer, dtcorrect
from .mapviewer import MapViewer, mapviewer


_larch_builtins = {'_plotter': {'dtcorrect_viewer': dtcorrect,
                                'map_viewer': mapviewer}}

# def initializeLarchPlugin(_larch=None):
#     """add MapFrameViewer to _sys.gui_apps """
#     if _larch is not None:
#         _sys = _larch.symtable._sys
#         if not hasattr(_sys, 'gui_apps'):
#             _sys.gui_apps = {}
#         _sys.gui_apps['mapviewer'] = ('XRF Map Viewer', MapViewerFrame)



#!/usr/bin/env python
'''
   Plotting functions for Larch, wrapping the mplot plotting
   widgets which use matplotlib

Exposed functions here are
   plot:  display 2D line plot to an enhanced,
            configurable Plot Frame
   oplot: overplot a 2D line plot on an existing Plot Frame
   imshow: display a false-color map from array data on
           a configurable Image Display Frame.
'''

import time
import os
import sys
import wx

from wxmplot import PlotFrame, ImageFrame, StackedPlotFrame

import larch
from ..xrf import isLarchMCAGroup
from ..larchlib import ensuremod
from ..site_config import usr_larchdir

from .xrfdisplay import XRFDisplayFrame

mpl_dir = os.path.join(usr_larchdir, 'matplotlib')
os.environ['MPLCONFIGDIR'] = mpl_dir
if not os.path.exists(mpl_dir):
    try:
        os.makedirs(mpl_dir)
    except:
        pass

from matplotlib.axes import Axes
HIST_DOC = Axes.hist.__doc__

IMG_DISPLAYS = {}
PLOT_DISPLAYS = {}
FITPLOT_DISPLAYS = {}
XRF_DISPLAYS = {}

_larch_name = '_plotter'

__DOC__ = '''
General Plotting and Image Display Functions

The functions here include (but are not limited to):

function         description
------------     ------------------------------
plot             2D (x, y) plotting, with many, many options
plot_text        add text to a 2D plot
plot_marker      add a marker to a 2D plot
plot_arrow       add an arrow to a 2D plot

imshow           image display (false-color intensity image)

xrf_plot         browsable display for XRF spectra
'''

MAX_WINDOWS = 20
MAX_CURSHIST = 25

class XRFDisplay(XRFDisplayFrame):
    def __init__(self, wxparent=None, window=1, _larch=None,
                 size=(725, 425), **kws):
        XRFDisplayFrame.__init__(self, parent=wxparent, size=size,
                                 _larch=_larch,
                                 exit_callback=self.onExit, **kws)
        self.Show()
        self.Raise()
        self.panel.cursor_callback = self.onCursor
        self.window = int(window)
        self._larch = _larch
        self._xylims = {}
        self.symname = '%s.xrf%i' % (_larch_name, self.window)
        symtable = ensuremod(self._larch, _larch_name)

        if symtable is not None:
            symtable.set_symbol(self.symname, self)
        if window not in XRF_DISPLAYS:
            XRF_DISPLAYS[window] = self

    def onExit(self, o, **kw):
        try:
            symtable = self._larch.symtable
            if symtable.has_group(_larch_name):
                symtable.del_symbol(self.symname)
        except:
            pass
        if self.window in XRF_DISPLAYS:
            XRF_DISPLAYS.pop(self.window)

        self.Destroy()

    def onCursor(self, x=None, y=None, **kw):
        symtable = ensuremod(self._larch, _larch_name)
        if symtable is None:
            return
        symtable.set_symbol('%s_xrf_x'  % self.symname, x)
        symtable.set_symbol('%s_xrf_y'  % self.symname, y)

class PlotDisplay(PlotFrame):
    def __init__(self, wxparent=None, window=1, _larch=None, size=None, **kws):
        PlotFrame.__init__(self, parent=None, size=size,
                           output_title='plot2d',
                           exit_callback=self.onExit, **kws)

        self.Show()
        self.Raise()
        self.panel.cursor_callback = self.onCursor
        self.panel.cursor_mode = 'zoom'
        self.window = int(window)
        self._larch = _larch
        self._xylims = {}
        self.cursor_hist = []
        self.symname = '%s.plot%i' % (_larch_name, self.window)
        symtable = ensuremod(self._larch, _larch_name)
        self.panel.canvas.figure.set_facecolor('#FDFDFB')

        if symtable is not None:
            symtable.set_symbol(self.symname, self)
            if not hasattr(symtable, '%s.cursor_maxhistory' % _larch_name):
                symtable.set_symbol('%s.cursor_maxhistory' % _larch_name, MAX_CURSHIST)

        if window not in PLOT_DISPLAYS:
            PLOT_DISPLAYS[window] = self

    def onExit(self, o, **kw):
        try:
            symtable = self._larch.symtable
            if symtable.has_group(_larch_name):
                symtable.del_symbol(self.symname)
        except:
            pass
        if self.window in PLOT_DISPLAYS:
            PLOT_DISPLAYS.pop(self.window)

        self.Destroy()

    def onCursor(self, x=None, y=None, **kw):
        symtable = ensuremod(self._larch, _larch_name)
        if symtable is None:
            return
        hmax = getattr(symtable, '%s.cursor_maxhistory' % _larch_name, MAX_CURSHIST)
        symtable.set_symbol('%s_x'  % self.symname, x)
        symtable.set_symbol('%s_y'  % self.symname, y)
        self.cursor_hist.insert(0, (x, y, time.time()))
        if len(self.cursor_hist) > hmax:
            self.cursor_hist = self.cursor_hist[:hmax]
        symtable.set_symbol('%s_cursor_hist' % self.symname, self.cursor_hist)


class StackedPlotDisplay(StackedPlotFrame):
    def __init__(self, wxparent=None, window=1, _larch=None,  size=None, **kws):
        StackedPlotFrame.__init__(self, parent=None,
                                  exit_callback=self.onExit, **kws)

        self.Show()
        self.Raise()
        self.panel.cursor_callback = self.onCursor
        self.panel.cursor_mode = 'zoom'
        self.window = int(window)
        self._larch = _larch
        self._xylims = {}
        self.cursor_hist = []
        self.symname = '%s.fitplot%i' % (_larch_name, self.window)
        symtable = ensuremod(self._larch, _larch_name)
        self.panel.canvas.figure.set_facecolor('#FDFDFB')
        self.panel_bot.canvas.figure.set_facecolor('#FDFDFB')

        if symtable is not None:
            symtable.set_symbol(self.symname, self)
            if not hasattr(symtable, '%s.cursor_maxhistory' % _larch_name):
                symtable.set_symbol('%s.cursor_maxhistory' % _larch_name, MAX_CURSHIST)

        if window not in FITPLOT_DISPLAYS:
            FITPLOT_DISPLAYS[window] = self

    def onExit(self, o, **kw):
        try:
            symtable = self._larch.symtable
            if symtable.has_group(_larch_name):
                symtable.del_symbol(self.symname)
        except:
            pass
        if self.window in FITPLOT_DISPLAYS:
            FITPLOT_DISPLAYS.pop(self.window)

        self.Destroy()

    def onCursor(self, x=None, y=None, **kw):
        symtable = ensuremod(self._larch, _larch_name)
        if symtable is None:
            return
        hmax = getattr(symtable, '%s.cursor_maxhistory' % _larch_name, MAX_CURSHIST)
        symtable.set_symbol('%s_x'  % self.symname, x)
        symtable.set_symbol('%s_y'  % self.symname, y)
        self.cursor_hist.insert(0, (x, y, time.time()))
        if len(self.cursor_hist) > hmax:
            self.cursor_hist = self.cursor_hist[:hmax]
        symtable.set_symbol('%s_cursor_hist' % self.symname, self.cursor_hist)

class ImageDisplay(ImageFrame):
    def __init__(self, wxparent=None, window=1, _larch=None, size=None, **kws):
        ImageFrame.__init__(self, parent=None, size=size,
                                  exit_callback=self.onExit, **kws)
        self.Show()
        self.Raise()
        self.cursor_pos = []
        self.panel.cursor_callback = self.onCursor
        self.panel.contour_callback = self.onContour
        self.window = int(window)
        self.symname = '%s.img%i' % (_larch_name, self.window)
        self._larch = _larch
        symtable = ensuremod(self._larch, _larch_name)
        if symtable is not None:
            symtable.set_symbol(self.symname, self)
        if self.window not in IMG_DISPLAYS:
            IMG_DISPLAYS[self.window] = self

    def onContour(self, levels=None, **kws):
        symtable = ensuremod(self._larch, _larch_name)
        if symtable is not None and levels is not None:
            symtable.set_symbol('%s_contour_levels'  % self.symname, levels)

    def onExit(self, o, **kw):
        try:
            symtable = self._larch.symtable
            symtable.has_group(_larch_name), self.symname
            if symtable.has_group(_larch_name):
                symtable.del_symbol(self.symname)
        except:
            pass
        if self.window in IMG_DISPLAYS:
            IMG_DISPLAYS.pop(self.window)
        self.Destroy()

    def onCursor(self,x=None, y=None, ix=None, iy=None, val=None, **kw):
        symtable = ensuremod(self._larch, _larch_name)
        if symtable is None:
            return
        set = symtable.set_symbol
        if x is not None:   set('%s_x' % self.symname, x)
        if y is not None:   set('%s_y' % self.symname, y)
        if ix is not None:  set('%s_ix' % self.symname, ix)
        if iy is not None:  set('%s_iy' % self.symname, iy)
        if val is not None: set('%s_val' % self.symname, val)

def _getDisplay(win=1, _larch=None, wxparent=None, size=None,
                wintitle=None, xrf=False, image=False, stacked=False):
    """make a plotter"""
    # global PLOT_DISPLAYS, IMG_DISPlAYS
    if  hasattr(_larch, 'symtable'):
        if (getattr(_larch.symtable._sys.wx, 'wxapp', None) is None or
            getattr(_larch.symtable._plotter, 'no_plotting', False)):
            return None
    win = max(1, min(MAX_WINDOWS, int(abs(win))))
    title   = 'Plot Window %i' % win
    symname = '%s.plot%i' % (_larch_name, win)
    creator = PlotDisplay
    display_dict = PLOT_DISPLAYS
    if image:
        creator = ImageDisplay
        display_dict = IMG_DISPLAYS
        title   = 'Image Window %i' % win
        symname = '%s.img%i' % (_larch_name, win)
    elif xrf:
        creator = XRFDisplay
        display_dict = XRF_DISPLAYS
        title   = 'XRF Display Window %i' % win
        symname = '%s.xrf%i' % (_larch_name, win)
    elif stacked:
        creator = StackedPlotDisplay
        display_dict = FITPLOT_DISPLAYS
        title   = 'Fit Plot Window %i' % win
        symname = '%s.fitplot%i' % (_larch_name, win)

    if wintitle is not None:
        title = wintitle

    def _get_disp(syname, creator, win, ddict, wxparent, size, _larch):
        if win in ddict:
            display = ddict[win]
        else:
            display = None
            if  hasattr(_larch, 'symtable'):
                display = _larch.symtable.get_symbol(symname, create=True)
            if display is None:
                display = creator(window=win, wxparent=wxparent,
                                  size=size, _larch=_larch)
            ddict[win] = display
        return display

    display = _get_disp(symname, creator, win, display_dict, wxparent,
                        size, _larch)
    try:
        display.SetTitle(title)
    except:
        display_dict.pop(win)
        display = _get_disp(symname, creator, win, display_dict, wxparent,
                            size, _larch)
        display.SetTitle(title) 
    if  hasattr(_larch, 'symtable'):
        _larch.symtable.set_symbol(symname, display)
    return display

def _xrf_plot(x=None, y=None, mca=None, win=1, new=True, as_mca2=False, _larch=None,
              wxparent=None, size=None, side='left', force_draw=True, wintitle=None,
              **kws):
    """xrf_plot(energy, data[, win=1], options])

    Show XRF trace of energy, data

    Parameters:
    --------------
        energy :  array of energies
        counts :  array of counts
        mca:      Group counting MCA data (rois, etc)
        as_mca2:  use mca as background MCA

        win: index of Plot Frame (0, 1, etc).  May create a new Plot Frame.
        new: flag (True/False, default False) for whether to start a new plot.
        color: color for trace (name such as 'red', or '#RRGGBB' hex string)
        style: trace linestyle (one of 'solid', 'dashed', 'dotted', 'dot-dash')
        linewidth:  integer width of line
        marker:  symbol to draw at each point ('+', 'o', 'x', 'square', etc)
        markersize: integer size of marker

    See Also: xrf_oplot, plot
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size,
                          _larch=_larch, wintitle=wintitle, xrf=True)
    if plotter is None:
        return
    plotter.Raise()
    if x is None:
        return

    if isLarchMCAGroup(x):
        mca = x
        y = x.counts
        x = x.energy

    if as_mca2:
        if isLarchMCAGroup(mca):
            plotter.add_mca(mca, as_mca2=True, plot=False)
            plotter.plotmca(mca, as_mca2=True, **kws)
        elif y is not None:
            plotter.oplot(x, y, mca=mca, as_mca2=True, **kws)
    elif new:
        if isLarchMCAGroup(mca):
            plotter.add_mca(mca, plot=False)
            plotter.plotmca(mca, **kws)
        elif y is not None:
            plotter.plot(x, y, mca=mca, **kws)
    elif y is not None:
        if isLarchMCAGroup(mca):
            plotter.add_mca(mca, plot=False)
        plotter.oplot(x, y, mca=mca, **kws)


def _xrf_oplot(x=None, y=None, mca=None, win=1, _larch=None, **kws):
    """xrf_oplot(energy, data[, win=1], options])

    Overplot a second  XRF trace of energy, data

    Parameters:
    --------------
        energy :  array of energies
        counts :  array of counts
        mca:      Group counting MCA data (rois, etc)

        win: index of Plot Frame (0, 1, etc).  May create a new Plot Frame.
        color: color for trace (name such as 'red', or '#RRGGBB' hex string)
        style: trace linestyle (one of 'solid', 'dashed', 'dotted', 'dot-dash')

    See Also: xrf_plot
    """
    _xrf_plot(x=x, y=y, mca=mca, win=win, _larch=_larch, new=False, **kws)

def _plot(x,y, win=1, new=False, _larch=None, wxparent=None, size=None,
          xrf=False, stacked=False, force_draw=True, side='left', wintitle=None, **kws):
    """plot(x, y[, win=1], options])

    Plot 2-D trace of x, y arrays in a Plot Frame, clearing any plot currently in the Plot Frame.

    Parameters:
    --------------
        x :  array of ordinate values
        y :  array of abscissa values (x and y must be same size!)

        win: index of Plot Frame (0, 1, etc).  May create a new Plot Frame.
        new: flag (True/False, default False) for whether to start a new plot.
        force_draw: flag (True/False, default Tree) for whether force a draw.
                    This will take a little extra time, and is not needed when
                    typing at the command-line, but is needed for plots to update
                    from inside scripts.
        label: label for trace
        title:  title for Plot
        xlabel: x-axis label
        ylabel: y-axis label
        ylog_scale: whether to show y-axis as log-scale (True or False)
        grid: whether to draw background grid (True or False)

        color: color for trace (name such as 'red', or '#RRGGBB' hex string)
        style: trace linestyle (one of 'solid', 'dashed', 'dotted', 'dot-dash')
        linewidth:  integer width of line
        marker:  symbol to draw at each point ('+', 'o', 'x', 'square', etc)
        markersize: integer size of marker

        drawstyle: style for joining line segments

        dy: array for error bars in y (must be same size as y!)
        yaxis='left'??
        use_dates

    See Also: oplot, newplot
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size,
                          xrf=xrf, stacked=stacked,
                          wintitle=wintitle,  _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    if new:
        plotter.plot(x, y, side=side, **kws)
    else:
        plotter.oplot(x, y, side=side, **kws)
    if force_draw:
        wx_update(_larch=_larch)

def _redraw_plot(win=1, xrf=False, stacked=False, size=None, wintitle=None,
                 _larch=None, wxparent=None):
    """redraw_plot(win=1)

    redraw a plot window, especially convenient to force setting limits after
    multiple plot()s with delay_draw=True
    """

    plotter = _getDisplay(wxparent=wxparent, win=win, size=size,
                          xrf=xrf, stacked=stacked,
                          wintitle=wintitle,  _larch=_larch)
    plotter.panel.unzoom_all()


def _update_trace(x, y, trace=1, win=1, _larch=None, wxparent=None,
                 side='left', redraw=False, **kws):
    """update a plot trace with new data, avoiding complete redraw"""
    plotter = _getDisplay(wxparent=wxparent, win=win, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    trace -= 1 # wxmplot counts traces from 0

    plotter.panel.update_line(trace, x, y, draw=True, side=side)
    wx_update(_larch=_larch)

def wx_update(_larch=None, **kws):
    if  hasattr(_larch, 'symtable'):
        _larch.symtable.set_symbol('_sys.wx.force_wxupdate', True)
    try:
        _larch.symtable.get_symbol('_sys.wx.ping')(timeout=0.002)
    except:
        pass

def _plot_setlimits(xmin=None, xmax=None, ymin=None, ymax=None, win=1, wxparent=None,
                    _larch=None):
    """set plot view limits for plot in window `win`"""
    plotter = _getDisplay(wxparent=wxparent, win=win, _larch=_larch)
    if plotter is None:
        return
    plotter.panel.set_xylims((xmin, xmax, ymin, ymax))

def _oplot(x, y, win=1, _larch=None, wxparent=None, xrf=False, stacked=False,
           size=None, **kws):
    """oplot(x, y[, win=1[, options]])

    Plot 2-D trace of x, y arrays in a Plot Frame, over-plotting any
    plot currently in the Plot Frame.

    This is equivalent to
    plot(x, y[, win=1[, new=False[, options]]])

    See Also: plot, newplot
    """
    kws['new'] = False
    _plot(x, y, win=win, size=size, xrf=xrf, stacked=stacked,
          wxparent=wxparent, _larch=_larch, **kws)

def _newplot(x, y, win=1, _larch=None, wxparent=None,  size=None, wintitle=None,
             **kws):
    """newplot(x, y[, win=1[, options]])

    Plot 2-D trace of x, y arrays in a Plot Frame, clearing any
    plot currently in the Plot Frame.

    This is equivalent to
    plot(x, y[, win=1[, new=True[, options]]])

    See Also: plot, oplot
    """
    _plot(x, y, win=win, size=size, new=True, _larch=_larch,
          wxparent=wxparent, wintitle=wintitle, **kws)

def _plot_text(text, x, y, win=1, side='left', size=None,
               stacked=False, xrf=False, rotation=None, ha='left', va='center',
               _larch=None, wxparent=None,  **kws):
    """plot_text(text, x, y, win=1, options)

    add text at x, y coordinates of a plot

    Parameters:
    --------------
        text:  text to draw
        x:     x position of text
        y:     y position of text
        win:   index of Plot Frame (0, 1, etc).  May create a new Plot Frame.
        side:  which axis to use ('left' or 'right') for coordinates.
        rotation:  text rotation. angle in degrees or 'vertical' or 'horizontal'
        ha:    horizontal alignment ('left', 'center', 'right')
        va:    vertical alignment ('top', 'center', 'bottom', 'baseline')

    See Also: plot, oplot, plot_arrow
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, xrf=xrf,
                          stacked=stacked, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()

    plotter.add_text(text, x, y, side=side,
                     rotation=rotation, ha=ha, va=va, **kws)

def _plot_arrow(x1, y1, x2, y2, win=1, side='left',
                shape='full', color='black',
                width=0.00, head_width=0.05, head_length=0.25,
               _larch=None, wxparent=None, stacked=False, xrf=False,
                size=None, **kws):

    """plot_arrow(x1, y1, x2, y2, win=1, **kws)

    draw arrow from x1, y1 to x2, y2.

    Parameters:
    --------------
        x1: starting x coordinate
        y1: starting y coordinate
        x2: ending x coordinate
        y2: ending y coordinate
        side: which axis to use ('left' or 'right') for coordinates.
        shape:  arrow head shape ('full', 'left', 'right')
        color:  arrow color ('black')
        width:  width of arrow line (in points. default=0.0)
        head_width:  width of arrow head (in points. default=0.05)
        head_length:  length of arrow head (in points. default=0.25)
        overhang:    amount the arrow is swept back (in points. default=0)
        win:  window to draw too

    See Also: plot, oplot, plot_text
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, xrf=xrf,
                          stacked=stacked, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    plotter.add_arrow(x1, y1, x2, y2, side=side, shape=shape,
                      color=color, width=width, head_length=head_length,
                      head_width=head_width, **kws)

def _plot_marker(x, y, marker='o', size=4, color='black', label='_nolegend_',
               _larch=None, wxparent=None, win=1, xrf=False, stacked=False, **kws):

    """plot_marker(x, y, marker='o', size=4, color='black')

    draw a marker at x, y

    Parameters:
    -----------
        x:      x coordinate
        y:      y coordinate
        marker: symbol to draw at each point ('+', 'o', 'x', 'square', etc) ['o']
        size:   symbol size [4]
        color:  color  ['black']

    See Also: plot, oplot, plot_text
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=None, xrf=xrf,
                          stacked=stacked, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    plotter.oplot([x], [y], marker=marker, markersize=size, label=label,
                 color=color, _larch=_larch, wxparent=wxparent,  **kws)

def _plot_axhline(y, xmin=0, xmax=1, win=1, wxparent=None, xrf=False,
                  stacked=False, size=None, delay_draw=False, _larch=None, **kws):
    """plot_axhline(y, xmin=None, ymin=None, **kws)

    plot a horizontal line spanning the plot axes
    Parameters:
    --------------
        y:      y position of line
        xmin:   starting x fraction (window units -- not user units!)
        xmax:   ending x fraction (window units -- not user units!)
    See Also: plot, oplot, plot_arrow
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, xrf=xrf,
                          stacked=stacked, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    if 'label' not in kws:
        kws['label'] = '_nolegend_'
    plotter.panel.axes.axhline(y, xmin=xmin, xmax=xmax, **kws)
    if delay_draw:
        plotter.panel.canvas.draw()

def _plot_axvline(x, ymin=0, ymax=1, win=1, wxparent=None, xrf=False,
                  stacked=False, size=None, delay_draw=False, _larch=None, **kws):
    """plot_axvline(y, xmin=None, ymin=None, **kws)

    plot a vertical line spanning the plot axes
    Parameters:
    --------------
        x:      x position of line
        ymin:   starting y fraction (window units -- not user units!)
        ymax:   ending y fraction (window units -- not user units!)
    See Also: plot, oplot, plot_arrow
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, xrf=xrf,
                          stacked=stacked, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    if 'label' not in kws:
        kws['label'] = '_nolegend_'
    plotter.panel.axes.axvline(x, ymin=ymin, ymax=ymax, **kws)
    if not delay_draw:
        plotter.panel.canvas.draw()

def _getcursor(win=1, timeout=30, _larch=None, wxparent=None, size=None,
               xrf=False, stacked=False, **kws):
    """get_cursor(win=1, timeout=30)

    waits (up to timeout) for cursor click in selected plot window, and
    returns x, y position of cursor.  On timeout, returns the last known
    cursor position, or (None, None)

    Note that _plotter.plotWIN_x and _plotter.plotWIN_y will be updated,
    with each cursor click, and so can be used to read the last cursor
    position without blocking.

    For a more consistent programmatic approach, this routine can be called
    with timeout <= 0 to read the most recently clicked cursor position.
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, xrf=xrf,
                          stacked=stacked, _larch=_larch)
    if plotter is None:
        return
    symtable = ensuremod(_larch, _larch_name)
    sentinal = '%s.plot%i_cursorflag' % (_larch_name, win)
    xsym = '%s.plot%i_x' % (_larch_name, win)
    ysym = '%s.plot%i_y' % (_larch_name, win)

    xval = symtable.get_symbol(xsym, create=True)
    yval = symtable.get_symbol(ysym, create=True)
    symtable.set_symbol(sentinal, False)

    def onChange(symbolname=None, **kws):
        symtable.set_symbol(kws['sentinal'], True)

    symtable.add_callback(xsym, onChange, kws={'sentinal': sentinal})

    t0 = time.time()
    while time.time() - t0 < timeout:
        wx_update(_larch=_larch)
        if symtable.get_symbol(sentinal):
            break
    symtable.del_symbol(sentinal)
    symtable.clear_callbacks(xsym)
    return (symtable.get_symbol(xsym), symtable.get_symbol(ysym))

def last_cursor_pos(win=None, _larch=None):
    """return most recent cursor position -- 'last click on plot'

    By default, this returns the last postion for all plot windows.
    If win is not `None`, the last position for that window will be returned

    Arguments
    ---------
    win  (int or None) index of window to get cursor position [None, all windows]

    Returns
    -------
    x, y coordinates of most recent cursor click, in user units
    """
    if  hasattr(_larch, 'symtable'):
        plotter = _larch.symtable._plotter
    else:
        return None, None
    histories = []
    for attr in dir(plotter):
        if attr.endswith('_cursor_hist'):
            histories.append(attr)

    if win is not None:
        tmp = []
        for attr in histories:
            if attr.startswith('plot%d_' % win):
                tmp.append(attr)
        histories = tmp
    _x, _y, _t = None, None, 0
    for hist in histories:
        for px, py, pt in getattr(plotter, hist, [None, None, -1]):
            if pt > _t and px is not None:
                _x, _y, _t = px, py, pt
    return _x, _y


def _scatterplot(x,y, win=1, _larch=None, wxparent=None, size=None,
          force_draw=True,  **kws):
    """scatterplot(x, y[, win=1], options])

    Plot x, y values as a scatterplot.  Parameters are very similar to
    those of plot()

    See Also: plot, newplot
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    plotter.scatterplot(x, y, **kws)
    if force_draw:
        wx_update(_larch=_larch)


def _fitplot(x, y, y2=None, panel='top', label=None, label2=None, win=1,
             _larch=None, wxparent=None, size=None, **kws):
    """fit_plot(x, y, y2=None, win=1, options)

    Plot x, y values in the top of a StackedPlot. If y2 is not None, then x, y2 values
    will also be plotted in the top frame, and the residual (y-y2) in the bottom panel.

    By default, arrays will be plotted in the top panel, and you must
    specify `panel='bot'` to plot an array in the bottom panel.

    Parameters are the same as for plot() and oplot()

    See Also: plot, newplot
    """
    plotter = _getDisplay(wxparent=wxparent, win=win, size=size,
                          stacked=True, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    plotter.plot(x, y, panel='top', label=label, **kws)
    if y2 is not None:
        kws.update({'label': label2})
        plotter.oplot(x, y2, panel='top', **kws)
        plotter.plot(x, y2-y, panel='bot')
        plotter.panel.conf.set_margins(top=0.15, bottom=0.01,
                                       left=0.15, right=0.05)
        plotter.panel_bot.conf.set_margins(top=0.01, bottom=0.35,
                                           left=0.15, right=0.05)


def _hist(x, bins=10, win=1, new=False,
           _larch=None, wxparent=None, size=None, force_draw=True,  *args, **kws):

    plotter = _getDisplay(wxparent=wxparent, win=win, size=size, _larch=_larch)
    if plotter is None:
        return
    plotter.Raise()
    if new:
        plotter.panel.axes.clear()

    out = plotter.panel.axes.hist(x, bins=bins, **kws)
    plotter.panel.canvas.draw()
    if force_draw:
        wx_update(_larch=_larch)
    return out


_hist.__doc__ = """
    hist(x, bins, win=1, options)

  %s
""" % (HIST_DOC)


def _imshow(map, x=None, y=None, colormap=None, win=1, _larch=None,
            wxparent=None, size=None, **kws):
    """imshow(map[, options])

    Display an 2-D array of intensities as a false-color map

    map: 2-dimensional array for map
    """
    img = _getDisplay(wxparent=wxparent, win=win, size=size, _larch=_larch, image=True)
    if img is not None:
        img.display(map, x=x, y=y, colormap=colormap, **kws)

def _contour(map, x=None, y=None, _larch=None, **kws):
    """contour(map[, options])

    Display an 2-D array of intensities as a contour plot

    map: 2-dimensional array for map
    """
    kws.update(dict(style='contour'))
    _imshow(map, x=x, y=y, _larch=_larch, **kws)

def _saveplot(fname, dpi=300, format=None, win=1, _larch=None, wxparent=None,
              size=None, facecolor='w', edgecolor='w', quality=90,
              image=False, **kws):
    """formats: png (default), svg, pdf, jpeg, tiff"""
    thisdir = os.path.abspath(os.curdir)
    if format is None:
        pref, suffix = os.path.splitext(fname)
        if suffix is not None:
            if suffix.startswith('.'):
                suffix = suffix[1:]
            format = suffix
    if format is None: format = 'png'
    format = format.lower()
    canvas = _getDisplay(wxparent=wxparent, win=win, size=size,
                         _larch=_larch, image=image).panel.canvas
    if canvas is None:
        return
    if format in ('jpeg', 'jpg'):
        canvas.print_jpeg(fname, quality=quality, **kws)
    elif format in ('tiff', 'tif'):
        canvas.print_tiff(fname, **kws)
    elif format in ('png', 'svg', 'pdf', 'emf', 'eps'):
        canvas.print_figure(fname, dpi=dpi, format=format,
                            facecolor=facecolor, edgecolor=edgecolor, **kws)
    else:
        print('unsupported image format: ', format)
    os.chdir(thisdir)

def _saveimg(fname, _larch=None, **kws):
    """save image from image display"""
    kws.update({'image':True})
    _saveplot(fname, _larch=_larch, **kws)

def _closeDisplays(_larch=None, **kws):
    for display in (PLOT_DISPLAYS, IMG_DISPLAYS,
                    FITPLOT_DISPLAYS, XRF_DISPLAYS):
        for win in display.values():
            win.Destroy()



#!/usr/bin/env pythonw
'''
GUI for displaying 2D XRD images

'''
import os
import numpy as np

import matplotlib.cm as colormap
from functools import partial

import h5py

import wx
try:
    from wx._core import PyDeadObjectError
except:
    PyDeadObjectError = Exception

from wxmplot import PlotPanel
from wxmplot.imagepanel import ImagePanel
from wxutils import MenuItem
from wxmplot.imageconf import ImageConfig,ColorMap_List

# import matplotlib.pyplot as plt

import larch
from larch import Group
from larch.larchlib import read_workdir, save_workdir
from larch.utils.strutils import bytes2str

from larch.io import tifffile, nativepath
from larch.xrd import (integrate_xrd,E_from_lambda,xrd1d,read_lambda,
                       calc_cake,twth_from_q,twth_from_d,
                       return_ai,twth_from_xy,q_from_xy,eta_from_xy)
from .XRDCalibrationFrame import CalibrationPopup
from .XRDMaskFrame import MaskToolsPopup
from .XRD1Dviewer import Calc1DPopup

###################################

VERSION = '2 (14-March-2018)'
SLIDER_SCALE = 1000. ## sliders step in unit 1. this scales to 0.001
PIXELS = 1024 #2048

QSTPS = 5000

XLABEL = ['q','2th','d']
XUNIT  = [u'q (\u212B\u207B\u00B9)',u'2\u03B8 (\u00B0)',u'd (\u212B)']

###################################

class diFFit2DPanel(wx.Panel):
    '''
    Panel for housing 2D XRD Image
    '''
    label='2D XRD'
    def __init__(self,parent,type='2D image',owner=None,_larch=None,size=(500, 500)):

        wx.Panel.__init__(self, parent)
        self.owner = owner
        self.type  = type
        self.ai    = None
        self.on_calibration()

        vbox = wx.BoxSizer(wx.VERTICAL)
        self.plot2D = ImagePanel(self,size=size,messenger=self.owner.write_message)
        vbox.Add(self.plot2D,proportion=1,flag=wx.ALL|wx.EXPAND,border = 10)
        self.SetSizer(vbox)

        self.plot2D.cursor_callback = self.on_cursor


    def on_calibration(self):

        if self.owner.calfile is not None and os.path.exists(self.owner.calfile):
            self.ai = return_ai(self.owner.calfile)

            self.ai.detector.shape = np.shape(self.owner.plt_img)

            fit2d_param = self.ai.getFit2D()
            self.center = (fit2d_param['centerX'],fit2d_param['centerY'])

    def on_cursor(self,x=None, y=None, **kw):

        if self.ai is not None:
            if self.type == 'cake':
                self.owner.twth = self.plot2D.xdata[int(x)]

                self.owner.xrd2Dcake.plot_line(x=x)
                self.owner.xrd2Dviewer.plot_ring()
            else:
                self.owner.twth = twth_from_xy(x,y,ai=self.ai)

                self.owner.xrd2Dviewer.plot_ring(x=x,y=y)
                self.owner.xrd2Dcake.plot_line()
        else:
            return

        self.owner.xrd1Dviewer.plot_line()

    def plot_line(self,x=None):

        if x is None: x = np.abs(self.plot2D.xdata-self.owner.twth).argmin()

        try:
            self.xrd_line.remove()
        except:
            pass

        self.xrd_line = self.plot2D.axes.axvline(x=x, color='r', linewidth=1)
        self.plot2D.canvas.draw()

    def plot_ring(self,x=None,y=None):
        import matplotlib.pyplot as plt
        if x is not None and y is not None:
            radius = np.sqrt((x-self.center[0])**2+(y-self.center[1])**2)
        else:
            radius = self.ai._dist * np.tan(np.radians(self.owner.twth))
            radius = radius / self.ai.detector.pixel1
        xrd_ring = plt.Circle(self.center, radius, color='red', fill=False)

        try:
            self.xrd_ring.remove()
        except:
            pass

        self.xrd_ring = self.plot2D.axes.add_artist(xrd_ring)
        self.plot2D.canvas.draw()

class diFFit1DPanel(wx.Panel):
    '''
    Panel for housing 1D XRD
    '''
    def __init__(self,parent,owner=None,_larch=None,size=(500, 100)):

        wx.Panel.__init__(self, parent)
        self.owner = owner

        vbox = wx.BoxSizer(wx.VERTICAL)
        self.plot1D = PlotPanel(self,size=size,messenger=self.owner.write_message)
        vbox.Add(self.plot1D,proportion=1,flag=wx.ALL|wx.EXPAND,border = 10)
        self.SetSizer(vbox)

        self.plot1D.cursor_callback = self.on_cursor

    def on_cursor(self,x=None, y=None, **kw):

        xi = self.owner.xaxis_type.GetSelection()
        if xi == 1:
            self.owner.twth = x
        else:
            ix = np.abs(self.owner.data1dxrd[1]-self.owner.twth).argmin()
            self.owner.twth = self.owner.data1dxrd[1][ix]

        self.owner.xrd1Dviewer.plot_line(x=x)
        self.owner.xrd2Dviewer.plot_ring()
        self.owner.xrd2Dcake.plot_line()

    def plot_line(self,x=None):

        if x is None:
            xi = self.owner.xaxis_type.GetSelection()
            if xi != 1:
                ix = np.abs(self.owner.data1dxrd[1]-self.owner.twth).argmin()
                x = self.owner.data1dxrd[xi][ix]
            else:
                x = self.owner.twth

        try:
            self.xrd_line.remove()
        except:
            pass

        self.xrd_line = self.plot1D.axes.axvline(x=x, color='r', linewidth=1)
        self.plot1D.draw()



class XRD2DViewerFrame(wx.Frame):
    '''
    Frame for housing all 2D XRD viewer widgets
    '''
    def __init__(self, _larch=None, xrd1Dviewer=None, ponifile=None, flip='vertical',
                 *args, **kw):

        screenSize = wx.DisplaySize()
        x,y = 1000,720 #1000,760
        if x > screenSize[0] * 0.9:
            x = int(screenSize[0] * 0.9)
            y = int(x*0.6)

        label = 'diFFit : 2D XRD Data Analysis Software'
        wx.Frame.__init__(self, None,title=label, size=(x,y))

        self.SetMinSize((700,500))

        self.statusbar = self.CreateStatusBar(3,wx.CAPTION)

        self.open_image = []
        self.open_scale = []

        ## Default image information
        self.raw_img = np.zeros((PIXELS,PIXELS))
        self.flp_img = np.zeros((PIXELS,PIXELS))
        self.plt_img = np.zeros((PIXELS,PIXELS))
        self.cake    = None
        self.twth    = None

        self.calfile = ponifile

        self.msk_img  = np.ones((PIXELS,PIXELS))
        self.bkgd_img = np.zeros((PIXELS,PIXELS))

        self.bkgd_scale = 0
        self.bkgdMAX = 2

        self.use_mask = False
        self.use_bkgd = False

        self.xrddisplay1D = xrd1Dviewer

        self.xrd1Dviewer = None
        self.xrd2Dcake  = None

        self.color = 'bone'
        self.flip = flip

        read_workdir('gsemap.dat')

        self.XRD2DMenuBar()
        self.Panel2DViewer()

        self.Centre()
        self.Show()

        if ponifile is None:
            self.btn_integ.Disable()
        else:
            self.btn_integ.Enable()

    def write_message(self, s, panel=0):
        '''write a message to the Status Bar'''
        self.SetStatusText(s, panel)

    def optionsON(self):

        if len(self.open_image) > 0:

            self.ch_clr.Enable()
            self.sldr_cntrst.Enable()
            self.entr_min.Enable()
            self.entr_max.Enable()
            self.btn_ct1.Enable()
            self.ch_flp.Enable()
            self.ch_scl.Enable()
            self.btn_mask.Enable()
            self.btn_bkgd.Enable()

            img_no = self.ch_img.GetSelection()
            if self.open_image[img_no].iframes > 1:
                self.hrz_frm_sldr.Enable()
                self.hrz_frm_sldr.SetRange(0,(self.open_image[img_no].iframes-1))
                self.hrz_frm_sldr.SetValue(self.open_image[img_no].i)
                for btn in self.hrz_frm_btn: btn.Enable()
            else:
                self.hrz_frm_sldr.Disable()
                self.hrz_frm_sldr.SetRange(0,0)
                self.hrz_frm_sldr.SetValue(0)
                for btn in self.hrz_frm_btn: btn.Disable()

            if self.open_image[img_no].jframes > 1:
                self.vrt_frm_sldr.Enable()
                self.vrt_frm_sldr.SetRange(0,(self.open_image[img_no].jframes-1))
                self.vrt_frm_sldr.SetValue(self.open_image[img_no].j)
                for btn in self.vrt_frm_btn: btn.Enable()
            else:
                self.vrt_frm_sldr.Disable()
                self.vrt_frm_sldr.SetRange(0,0)
                self.vrt_frm_sldr.SetValue(0)
                for btn in self.vrt_frm_btn: btn.Disable()

##############################################
#### OPENING AND DISPLAYING IMAGES

    def loadIMAGE(self,event=None):
        wildcards = '2DXRD image files (*.*)|*.*|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Choose 2D XRD image',
                           defaultDir=os.getcwd(),
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = '', False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            print('\nReading XRD image file: %s' % path)

            image,xrmfile = None,None
            try:
                xrmfile = h5py.File(path, 'r')
            except IOError:
                try:
                    from larch.io import read_xrd_netcdf
                    image = read_xrd_netcdf(path)
                except TypeError:
                    try:
                        image = tifffile.imread(path)
                    except ValueError:
                        print('Could not read file.')
                        return

            iname = os.path.split(path)[-1]
            self.plot2Dxrd(iname, image, path=path, h5file=xrmfile)

    def close2Dxrd(self,event=None):

        img_no = self.ch_img.GetSelection()
        try:
            print('Closing XRD image file: %s' % self.open_image[img_no].path)
            try:
                self.open_image[img_no].h5file.close()
            except:
                pass
            del self.open_image[img_no]
        except:
            pass

        self.ch_img.Set([image.label for image in self.open_image])
        if len(self.open_image) > 0:
            self.ch_img.SetSelection(0)
            self.selectIMAGE()
        else:
            self.clearIMAGE()


    def plot2Dxrd(self,iname,image,path='',h5file=None):

        self.write_message('Displaying image: %s' % iname, panel=0)

        try:
            new_img = XRDImg(label=iname, path=path, image=image, h5file=h5file)
        except:
            fail_msg = 'Image failed to load.'
            self.write_message(fail_msg, panel=0)
            print (fail_msg)
            return

        self.open_image.append(new_img)

        if self.open_image[-1].calfile is not None:
            if os.path.exists(self.open_image[-1].calfile):
                self.calfile = self.open_image[-1].calfile
                self.xrd2Dviewer.on_calibration()


        self.ch_img.Set([image.label for image in self.open_image])
        self.ch_img.SetStringSelection(iname)

        self.raw_img = self.open_image[-1].get_image()
        self.twth = None
        self.displayIMAGE(auto_contrast=True,unzoom=True)

        if self.open_image[-1].iframes > 1:
            self.hrz_frm_sldr.SetRange(0,(self.open_image[-1].iframes-1))
            self.hrz_frm_sldr.SetValue(self.open_image[-1].i)
        else:
            self.hrz_frm_sldr.Disable()
            for btn in self.hrz_frm_btn: btn.Disable()

        if self.open_image[-1].jframes > 1:
            self.vrt_frm_sldr.SetRange(0,(self.open_image[-1].jframes-1))
            self.vrt_frm_sldr.SetValue(self.open_image[-1].j)
        else:
            self.vrt_frm_sldr.Disable()
            for btn in self.vrt_frm_btn: btn.Disable()


    def changeFRAME(self,flag='hslider',event=None):

        img_no = self.ch_img.GetSelection()
        if self.open_image[img_no].iframes > 1 or self.open_image[img_no].jframes > 1:
            i,j = self.open_image[img_no].i,self.open_image[img_no].j
            if   flag=='next':     i = i + 1
            elif flag=='previous': i = i - 1
            elif flag=='hslider':  i = self.hrz_frm_sldr.GetValue()
            elif flag=='up':       j = j + 1
            elif flag=='down':     j = j - 1
            elif flag=='vslider':  j = self.vrt_frm_sldr.GetValue()

            if i < 0: i = self.open_image[img_no].iframes + i
            if j < 0: j = self.open_image[img_no].jframes + j
            if i > self.open_image[img_no].iframes: i = i - self.open_image[img_no].iframes
            if j > self.open_image[img_no].jframes: j = j - self.open_image[img_no].jframes

            self.raw_img =  self.open_image[img_no].get_image(i=i,j=j)

            self.hrz_frm_sldr.SetValue(i)
            self.vrt_frm_sldr.SetValue(j)
            self.displayIMAGE(auto_contrast=False,unzoom=False)#unzoom=True)

    def clearIMAGE(self):

        try:
            self.xrd2Dviewer.plot2D.clear()
            self.xrd2Dviewer.plot2D.redraw()
        except:
            pass

    def displayIMAGE(self,auto_contrast=True,unzoom=False):

        self.flipIMAGE()
        self.checkIMAGE()
        self.calcIMAGE()

        if unzoom:
            self.xrd2Dviewer.plot2D.display(self.plt_img,unzoom=unzoom)
        else:
            self.xrd2Dviewer.plot2D.conf.data = self.plt_img
            self.xrd2Dviewer.plot2D.redraw()
        self.display1DXRD()

        if auto_contrast: self.setContrast(auto_contrast=True)

        self.txt_ct2.SetLabel('[ image range: %i to %i ]' %
                         (np.min(self.plt_img),np.max(self.plt_img)))

        self.optionsON()
        self.xrd2Dviewer.plot2D.redraw()

    def redrawIMAGE(self,unzoom=False):

        self.flipIMAGE()
        self.checkIMAGE()
        self.calcIMAGE()
        self.colorIMAGE()

        self.xrd2Dviewer.plot2D.redraw()
        self.display1DXRD()

    def selectIMAGE(self,event=None):

        img_no = self.ch_img.GetSelection()
        self.raw_img = self.open_image[img_no].get_image()

        self.write_message('Displaying image: %s' % self.open_image[img_no].label, panel=0)
        try:
            self.displayIMAGE(auto_contrast=False,unzoom=True)
            self.setContrast()
        except:
            self.write_message('Image failed to load.', panel=0)

    def onChangeXscale(self,event=None):

        xi = self.xaxis_type.GetSelection()

        self.xrd1Dviewer.plot1D.update_line(0, self.data1dxrd[xi],self.data1dxrd[3])
        if self.twth is not None: self.xrd1Dviewer.plot_line()
        self.xrd1Dviewer.plot1D.set_viewlimits()
        self.xrd1Dviewer.plot1D.set_xlabel(XUNIT[xi])
        if xi == 2: self.xrd1Dviewer.plot1D.axes.set_xlim(np.min(self.data1dxrd[xi]), 6)
        self.xrd1Dviewer.plot1D.draw()


##############################################
#### IMAGE DISPLAY FUNCTIONS

    def calcIMAGE(self):
        if self.use_mask is True:
            if self.use_bkgd is True:
                self.plt_img = self.flp_img * self.msk_img - self.bkgd_img * self.bkgd_scale
            else:
                self.plt_img = self.flp_img * self.msk_img
        else:
            if self.use_bkgd is True:
                self.plt_img = self.flp_img - self.bkgd_img * self.bkgd_scale
            else:
                self.plt_img = self.flp_img

    def flipIMAGE(self):
        if self.flip == 'vertical': # Vertical
            self.flp_img = self.raw_img[::-1,:]
        elif self.flip == 'horizontal': # Horizontal
            self.flp_img = self.raw_img[:,::-1]
        elif self.flip == 'both': # both
            self.flp_img = self.raw_img[::-1,::-1]
        else: # None
            self.flp_img = self.raw_img

    def checkIMAGE(self):
        ## Reshapes/replaces mask and/or background if shape doesn't match that of image
        if self.msk_img.shape != self.raw_img.shape:
            self.msk_img = np.ones(self.raw_img.shape)
        if self.bkgd_img.shape != self.raw_img.shape:
            self.bkgd_img = np.zeros(self.raw_img.shape)

        ## Calculates the number of pixels in image, masked pixels, and background pixels
        img_pxls  = self.raw_img.shape[0]*self.raw_img.shape[1]
        msk_pxls  = img_pxls - int(sum(sum(self.msk_img)))
        bkgd_pxls = int(sum(sum(self.bkgd_img)))

        ## Enables mask checkbox.
        if msk_pxls == 0 or msk_pxls == img_pxls:
            self.ch_msk.Disable()
            self.msk_img = np.ones(self.raw_img.shape)
        else:
            self.ch_msk.Enable()

        ## Enables background slider and sets range.
        if bkgd_pxls == 0:
            self.sldr_bkgd.Disable()
            self.use_bkgd = False
        else:
            self.sldr_bkgd.Enable()
            self.use_bkgd = True

        self.sldr_bkgd.SetRange(0,self.bkgdMAX*SLIDER_SCALE)
        self.sldr_bkgd.SetValue(self.bkgd_scale*SLIDER_SCALE)

    def colorIMAGE(self,unzoom=False):
        self.xrd2Dviewer.plot2D.conf.cmap[0] = getattr(colormap, self.color)
        self.xrd2Dviewer.plot2D.display(self.plt_img) #,unzoom=unzoom)

        if self.cake is not None:
            self.xrd2Dcake.plot2D.conf.cmap[0] = getattr(colormap, self.color)
            self.xrd2Dcake.plot2D.display(self.cake[0]) #,unzoom=unzoom)

    def setCOLOR(self,event=None):
        if self.color != self.ch_clr.GetString(self.ch_clr.GetSelection()):
            self.color = self.ch_clr.GetString(self.ch_clr.GetSelection())
            self.colorIMAGE()

    def setFLIP(self,event=None):

        if self.flip != self.ch_flp.GetString(self.ch_flp.GetSelection()):
            self.flip = self.ch_flp.GetString(self.ch_flp.GetSelection())
            self.redrawIMAGE(unzoom=False)

    def setZSCALE(self,event=None):
        if self.ch_scl.GetSelection() == 1: ## log
            self.xrd2Dviewer.plot2D.conf.log_scale = True
            if self.xrd2Dcake is not None:
                self.xrd2Dcake.plot2D.conf.log_scale = True
            if self.xrd1Dviewer is not None:
                self.xrd1Dviewer.plot1D.axes.set_yscale('log')
                self.xrd1Dviewer.plot1D.set_viewlimits()
                self.xrd1Dviewer.plot1D.draw()
        else:  ## linear
            self.xrd2Dviewer.plot2D.conf.log_scale = False
            if self.xrd2Dcake is not None:
                self.xrd2Dcake.plot2D.conf.log_scale = False
            if self.xrd1Dviewer is not None:
                self.xrd1Dviewer.plot1D.axes.set_yscale('linear')
                self.xrd1Dviewer.plot1D.set_viewlimits()
                self.xrd1Dviewer.plot1D.draw()

        self.xrd2Dviewer.plot2D.redraw()
        if self.cake is not None:
            self.xrd2Dcake.plot2D.redraw()


##############################################
#### BACKGROUND FUNCTIONS
    def onBkgdScale(self,event=None):
        self.bkgd_scale = self.sldr_bkgd.GetValue()/SLIDER_SCALE
        self.redrawIMAGE(unzoom=False)

##############################################
#### IMAGE CONTRAST FUNCTIONS

    def onContrastRange(self,event=None):

        img_no = self.ch_img.GetSelection()
        img = self.open_image[img_no]

        img.minval = int(self.entr_min.GetValue())
        img.maxval = int(self.entr_max.GetValue())

        self.setContrast()


    def onSlider(self,event=None):

        img_no = self.ch_img.GetSelection()
        img = self.open_image[img_no]

        curval = int(self.sldr_cntrst.GetValue())

        self.xrd2Dviewer.plot2D.conf.auto_intensity = False
        self.xrd2Dviewer.plot2D.conf.int_lo[0] = img.minval
        self.xrd2Dviewer.plot2D.conf.int_hi[0] = curval
        self.xrd2Dviewer.plot2D.redraw()

        if self.cake is not None:
            self.xrd2Dcake.plot2D.conf.auto_intensity = False
            self.xrd2Dcake.plot2D.conf.int_lo[0] = img.minval
            self.xrd2Dcake.plot2D.conf.int_hi[0] = curval
            self.xrd2Dcake.plot2D.redraw()

    def setContrast(self,event=None,auto_contrast=False):
        img_no = self.ch_img.GetSelection()
        img = self.open_image[img_no]

        if auto_contrast: img.set_contrast(np.min(self.plt_img),np.max(self.plt_img))

        self.xrd2Dviewer.plot2D.conf.auto_intensity = False
        self.xrd2Dviewer.plot2D.conf.int_lo[0] = img.minval
        if auto_contrast:
            self.xrd2Dviewer.plot2D.conf.int_hi[0] = img.maxval*0.4
        else:
            self.xrd2Dviewer.plot2D.conf.int_hi[0] = img.maxval
        self.xrd2Dviewer.plot2D.redraw()

        self.sldr_cntrst.SetRange(img.minval,img.maxval)
        if auto_contrast:
            self.sldr_cntrst.SetValue(int(img.maxval*0.4))
        else:
            self.sldr_cntrst.SetValue(img.maxval)
        self.entr_min.SetValue('%i' % img.minval)
        self.entr_max.SetValue('%i' % img.maxval)

        if self.cake is not None:
            self.xrd2Dcake.plot2D.conf.auto_intensity = False
            self.xrd2Dcake.plot2D.conf.int_lo[0] = img.minval
            if auto_contrast:
                self.xrd2Dcake.plot2D.conf.int_hi[0] = img.maxval*0.4
            else:
                self.xrd2Dcake.plot2D.conf.int_hi[0] = img.maxval
            self.xrd2Dcake.plot2D.redraw()


##############################################
#### XRD MANIPULATION FUNTIONS

    def saveIMAGE(self,event=None,raw=False):
        wildcards = 'XRD image (*.tiff)|*.tiff|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, 'Save image as...',
                           defaultDir=os.getcwd(),
                           wildcard=wildcards,
                           style=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)

        path, save = None, False
        if dlg.ShowModal() == wx.ID_OK:
            save = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if save:
            if raw:
                tifffile.imsave(path,self.raw_img)
            else:
                tifffile.imsave(path,self.plt_img)

    def on1DXRD(self,event=None):

        read, save, plot = False, False, False
        if self.calfile is not None and self.plt_img is not None:
            myDlg = Calc1DPopup(self,self.plt_img)
            if myDlg.ShowModal() == wx.ID_OK:
                read = True
                save = myDlg.ch_save.GetValue()
                plot = myDlg.ch_plot.GetValue()
                unts = myDlg.save_choice.GetSelection()
                wdgs = myDlg.wedges.GetValue()
                if int(myDlg.xstep.GetValue()) < 1:
                    attrs = {'steps':5001}
                else:
                    attrs = {'steps':int(myDlg.xstep.GetValue())}
                unit = '2th' if unts == 1 else 'q'
                attrs.update({'unit':unit,'verbose':True})
            myDlg.Destroy()
        else:
            print('Data and calibration files must be available for this function.')
            return

        if read:
            if save:
                wildcards = '1D XRD file (*.xy)|*.xy|All files (*.*)|*.*'
                dlg = wx.FileDialog(self, 'Save file as...',
                                   defaultDir=os.getcwd(),
                                   wildcard=wildcards,
                                   style=wx.FD_SAVE|wx.FD_OVERWRITE_PROMPT)
                if dlg.ShowModal() == wx.ID_OK:
                    filename = dlg.GetPath().replace('\\', '/')
                    if not filename.endswith('.xy'):
                        filename = '%s.xy' % filename
                    attrs.update({'file':filename})

                dlg.Destroy()
            data1D = integrate_xrd(self.plt_img,self.calfile,**attrs)
            if wdgs > 1:
                xrdq_wdg,xrd1d_wdg,lmts_wdg = [],[],[]
                wdg_sz = 360./int(wdgs)
                for iwdg in range(wdgs):
                    wdg_lmts = np.array([iwdg*wdg_sz, (iwdg+1)*wdg_sz]) - 180
                    attrs.update({'wedge_limits':wdg_lmts})

                    if save:
                        wedgename = '%s_%i_to_%ideg.xy' % (filename.split('.xy')[0],
                                                           (wdg_lmts[0]+180),
                                                           (wdg_lmts[1]+180))
                        attrs.update({'file':wedgename})
                    q,counts = integrate_xrd(self.plt_img,self.calfile,**attrs)
                    xrdq_wdg  += [q]
                    xrd1d_wdg += [counts]
                    lmts_wdg  += [wdg_lmts]


            if plot:
                if self.xrddisplay1D is None:
                    self.xrddisplay1D = diFFit1DFrame()

                kwargs = {}
                wvlngth = read_lambda(self.calfile)
                kwargs.update({'wavelength':wvlngth,'energy':E_from_lambda(wvlngth)})
                kwargs.update({'label':self.open_image[self.ch_img.GetSelection()].label})
                data1dxrd = xrd1d(**kwargs)
                data1dxrd.xrd_from_2d(data1D,attrs['unit'])

                try:
                    self.xrddisplay1D.xrd1Dviewer.add1Ddata(data1dxrd)
                except PyDeadObjectError:
                    self.xrddisplay1D = diFFit1DFrame()
                    self.xrddisplay1D.xrd1Dviewer.add1Ddata(data1dxrd)

                if wdgs > 1:
                    for lmts,q,cnts in zip(lmts_wdg,xrdq_wdg,xrd1d_wdg):
                        label = '%s (%i to %i deg)' % (self.open_image[self.ch_img.GetSelection()].label,
                                                        (lmts[0]+180), (lmts[1]+180))
                        kwargs.update({'label':label})
                        data1dxrd = xrd1d(**kwargs)
                        data1dxrd.xrd_from_2d([q,cnts],'q')
                        self.xrddisplay1D.xrd1Dviewer.add1Ddata(data1dxrd)
                self.xrddisplay1D.Show()




##############################################
#### CALIBRATION FUNCTIONS
    def display1DXRD(self,event=None):

        if len(self.open_image) > 0 and self.calfile is not None:

            if self.xrd2Dcake is None or self.xrd1Dviewer is None:
                rightside = self.RightSidePanel(self.panel)
                self.panel2D.Add(rightside,proportion=1,flag=wx.EXPAND|wx.ALL,border=10)
                self.panel2D.Layout()
                self.Fit()
                self.SetSize((1400,720))

                tools1dxrd = self.ToolBox_1DXRD(self.panel)
                self.leftside.Add(tools1dxrd,flag=wx.ALL|wx.EXPAND,border=10)
                self.leftside.Layout()

                self.panel2D.Layout()
                self.Fit()
                self.SetSize((1400,720))

            self.btn_integ.Enable()

            ## Cake calculations
            xi = 1
            self.xaxis_type.SetSelection(xi)

            self.cake = calc_cake(self.plt_img, self.calfile, unit=XLABEL[xi], xsteps=QSTPS, ysteps=QSTPS)
            self.xrd2Dcake.plot2D.display(self.cake[0],x=self.cake[1],y=self.cake[2],
                                          xlabel=XLABEL[xi],ylabel='eta')
            self.xrd2Dcake.plot2D.conf.auto_intensity = False
            self.xrd2Dcake.plot2D.conf.int_lo[0] = self.xrd2Dviewer.plot2D.conf.int_lo[0]
            self.xrd2Dcake.plot2D.conf.int_hi[0] = self.xrd2Dviewer.plot2D.conf.int_hi[0]
            self.xrd2Dcake.plot2D.redraw()

            ## 1DXRD calculations

            q,I = integrate_xrd(self.plt_img, self.calfile, unit=XLABEL[xi], steps=QSTPS)
            self.data1dxrd = xrd1d(x=q,xtype=XLABEL[xi],I=I).all_data()

            self.xrd1Dviewer.plot1D.plot(self.data1dxrd[xi],self.data1dxrd[3],color='blue',
                                         xlabel=XUNIT[xi])
            self.xrd1Dviewer.plot1D.axes.yaxis.set_visible(False)
            self.xrd1Dviewer.plot1D.axes.spines['left'].set_visible(False)
            self.xrd1Dviewer.plot1D.axes.spines['right'].set_visible(False)
            self.xrd1Dviewer.plot1D.axes.spines['top'].set_visible(False)
            self.xrd1Dviewer.plot1D.draw()


    def Calibrate(self,event=None):

        CalibrationPopup(self)

    def openPONI(self,event=None):

        wildcards = 'pyFAI calibration file (*.poni)|*.poni|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Choose pyFAI calibration file',
                           defaultDir=os.getcwd(),
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            print('Loading calibration file:\n\t%s' % path)
            self.calfile = path
            self.xrd2Dviewer.on_calibration()
            self.display1DXRD()


##############################################
#### BACKGROUND FUNCTIONS
    def clearBkgd(self,event=None):

        self.bkgd = np.zeros(self.raw_img.shape)
        self.checkIMAGE()

    def openBkgd(self,event=None):

        wildcards = 'XRD background image (*.edf,*.tif,*.tiff)|*.tif;*.tiff;*.edf|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Choose XRD background image',
                           defaultDir=os.getcwd(),
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            try:
                self.bkgd_img = np.array(tifffile.imread(path))
                self.checkIMAGE()
                print('Reading background:\n\t%s' % path)
            except:
                print('\nCannot read as an image file: %s\n' % path)
                return


##############################################
#### MASK FUNCTIONS
    def openMask(self,event=None):

        wildcards = 'pyFAI mask file (*.edf)|*.edf|All files (*.*)|*.*'
        dlg = wx.FileDialog(self, message='Choose pyFAI mask file',
                           defaultDir=os.getcwd(),
                           wildcard=wildcards, style=wx.FD_OPEN)

        path, read = None, False
        if dlg.ShowModal() == wx.ID_OK:
            read = True
            path = dlg.GetPath().replace('\\', '/')
        dlg.Destroy()

        if read:
            try:
                try:
                    raw_mask = np.array(tifffile.imread(path))
                except:
                    import fabio
                    raw_mask = fabio.open(path).data
                self.msk_img = np.ones(raw_mask.shape)-raw_mask
                self.checkIMAGE()
                print('Reading mask:\n\t%s' % path)
            except:
                print('\nCannot read as mask file: %s\n' % path)
                return

        self.ch_msk.SetValue(True)
        self.applyMask(event=True)

#     def createMask(self,event=None):
#
#         MaskToolsPopup(self)
#         print('Popup to create mask!')

    def clearMask(self,event=None):

        self.msk_img = np.zeros(self.raw_img.shape)
        self.checkIMAGE()

    def applyMask(self,event=None):

        self.use_mask = self.ch_msk.GetValue()
        self.redrawIMAGE(unzoom=False)

##############################################
#### HELP FUNCTIONS
    def onAbout(self, event=None):
        info = wx.AboutDialogInfo()
        info.SetName('diFFit2D XRD Data Viewer')
        desc = 'Using X-ray Larch version: %s' % larch.version.__version__
        info.SetDescription(desc)
        info.SetVersion(VERSION)
        info.AddDeveloper('Margaret Koker: koker at cars.uchicago.edu')
        dlg = wx.AboutBox(info)

##############################################
#### MENU FUNCTIONS
    def onFolderSelect(self, evt=None):
        style = wx.DD_DIR_MUST_EXIST|wx.DD_DEFAULT_STYLE
        dlg = wx.DirDialog(self, 'Select Working Directory:', os.getcwd(),
                           style=style)

        if dlg.ShowModal() == wx.ID_OK:
            basedir = os.path.abspath(str(dlg.GetPath()))
            try:
                if len(basedir)  > 0:
                    os.chdir(nativepath(basedir))
                    save_workdir(nativepath(basedir))
            except OSError:
                print( 'Changed folder failed')
                pass
        save_workdir('gsemap.dat')
        dlg.Destroy()

    def onExit(self, event=None):

        dlg = wx.MessageDialog(None, 'Really Quit?', 'Question',
                               wx.YES_NO | wx.NO_DEFAULT | wx.ICON_QUESTION)

        ret = dlg.ShowModal()
        if ret != wx.ID_YES:
            return

        for image in self.open_image:
            try:
                image.h5file.close()
            except:
                pass

        try:
            self.Destroy()
        except:
            pass

##############################################
#### PANEL DEFINITIONS
    def XRD2DMenuBar(self):

        menubar = wx.MenuBar()

        ###########################
        ## diFFit2D
        diFFitMenu = wx.Menu()

        MenuItem(self, diFFitMenu, '&Open diffration image', '', self.loadIMAGE)
        MenuItem(self, diFFitMenu, 'Sa&ve displayed image to file', '', partial(self.saveIMAGE,raw=False))
        MenuItem(self, diFFitMenu, 'Save r&aw image to file', '', partial(self.saveIMAGE,raw=True))
        MenuItem(self, diFFitMenu, 'Change larch &working folder', '', self.onFolderSelect)

#         MenuItem(self, diFFitMenu, '&Save settings', '', None)
#         MenuItem(self, diFFitMenu, '&Load settings', '', None)
#         MenuItem(self, diFFitMenu, 'A&dd analysis to map file', '', None)
        MenuItem(self, diFFitMenu, '&Quit', 'Quit program', self.onExit)

        menubar.Append(diFFitMenu, '&diFFit2D')

        ###########################
        ## Process
        ProcessMenu = wx.Menu()

        MenuItem(self, ProcessMenu, 'Load &mask file', '', self.openMask)
        MenuItem(self, ProcessMenu, 'Remove current mas&k', '', self.clearMask)
#         MenuItem(self, ProcessMenu, 'C&reate mas&k', '', self.createMask)
        ProcessMenu.AppendSeparator()
        MenuItem(self, ProcessMenu, 'Load &background image', '', self.openBkgd)
        MenuItem(self, ProcessMenu, 'Remove current back&ground image', '', self.clearBkgd)
        ProcessMenu.AppendSeparator()
        MenuItem(self, ProcessMenu, 'Close current image', '', self.close2Dxrd)

        menubar.Append(ProcessMenu, '&Process')

        ###########################
        ## Analyze
        AnalyzeMenu = wx.Menu()

#         MenuItem(self, AnalyzeMenu, '&Calibrate', '', self.Calibrate)
        MenuItem(self, AnalyzeMenu, 'Load cali&bration file', '', self.openPONI)
#         MenuItem(self, AnalyzeMenu, 'Show current calibratio&n', '', None)
        AnalyzeMenu.AppendSeparator()
        MenuItem(self, AnalyzeMenu, '&Integrate (open 1D viewer)', '', self.on1DXRD)

        menubar.Append(AnalyzeMenu, 'Anal&yze')

        ###########################
        ## Help
        HelpMenu = wx.Menu()

        MenuItem(self, HelpMenu, '&About', 'About diFFit2D viewer', self.onAbout)

        menubar.Append(HelpMenu, '&Help')

        ###########################
        ## Create Menu Bar
        self.SetMenuBar(menubar)
        self.Bind(wx.EVT_CLOSE, self.onExit)

    def LeftSidePanel(self,panel):

        vbox = wx.BoxSizer(wx.VERTICAL)

        imgbox   = self.ImageBox(self.panel)
        vistools = self.Toolbox(self.panel)

        vbox.Add(imgbox,flag=wx.ALL|wx.EXPAND,border=10)
        vbox.Add(vistools,flag=wx.ALL|wx.EXPAND,border=10)

        return vbox

    def Panel2DViewer(self):
        '''
        Frame for housing all 2D XRD viewer widgets
        '''
        self.panel = wx.Panel(self)

        self.leftside = self.LeftSidePanel(self.panel)
        center = self.CenterPanel(self.panel)
        #rightside = self.RightSidePanel(self.panel)

        self.panel2D = wx.BoxSizer(wx.HORIZONTAL)
        self.panel2D.Add(self.leftside,flag=wx.ALL,border=10)
        self.panel2D.Add(center,proportion=1,flag=wx.EXPAND|wx.ALL,border=10)
        #self.panel2D.Add(rightside,proportion=1,flag=wx.EXPAND|wx.ALL,border=10)

        self.panel.SetSizer(self.panel2D)


    def ToolBox_1DXRD(self,panel):

        tlbx = wx.StaticBox(self.panel,label='1DXRD TOOLBOX')
        hbox = wx.StaticBoxSizer(tlbx,wx.HORIZONTAL)

        ttl_xaxis = wx.StaticText(self, label='X-SCALE')
        xunits = [u'q (\u212B\u207B\u00B9)',u'2\u03B8 (\u00B0)',u'd (\u212B)']

        self.xaxis_type = wx.Choice(self,choices=xunits)
        self.xaxis_type.Bind(wx.EVT_CHOICE, self.onChangeXscale)

        hbox.Add(ttl_xaxis, flag=wx.RIGHT, border=8)
        hbox.Add(self.xaxis_type, flag=wx.EXPAND, border=8)

        return hbox

    def ImageBox(self,panel):
        '''
        Frame for data toolbox
        '''

        tlbx = wx.StaticBox(self.panel,label='DISPLAYED IMAGE')
        vbox = wx.StaticBoxSizer(tlbx,wx.VERTICAL)


        ###########################
        ## DATA CHOICE


        self.ch_img = wx.Choice(self.panel,choices=[])
        self.ch_img.Bind(wx.EVT_CHOICE, self.selectIMAGE)
        vbox.Add(self.ch_img, flag=wx.EXPAND|wx.ALL, border=8)

        self.hrz_frm_sldr = wx.Slider(self.panel, minValue=0, maxValue=100, size=(120,-1),
                                 style = wx.SL_HORIZONTAL|wx.SL_LABELS)
        self.vrt_frm_sldr = wx.Slider(self.panel, minValue=0, maxValue=100, size=(-1,120),
                                 style = wx.SL_VERTICAL|wx.SL_LABELS|wx.SL_INVERSE)

        self.vrt_frm_btn = [ wx.Button(self.panel,label=u'\u2191', size=(40, -1)),
                             wx.Button(self.panel,label=u'\u2193', size=(40, -1))]
        self.hrz_frm_btn = [ wx.Button(self.panel,label=u'\u2190', size=(40, -1)),
                             wx.Button(self.panel,label=u'\u2192', size=(40, -1))]

        self.hrz_frm_btn[0].Bind(wx.EVT_BUTTON, partial(self.changeFRAME,'previous') )
        self.hrz_frm_btn[1].Bind(wx.EVT_BUTTON, partial(self.changeFRAME,'next')     )
        self.hrz_frm_sldr.Bind(wx.EVT_SLIDER,    partial(self.changeFRAME,'hslider')   )

        self.vrt_frm_btn[0].Bind(wx.EVT_BUTTON, partial(self.changeFRAME,'up') )
        self.vrt_frm_btn[1].Bind(wx.EVT_BUTTON, partial(self.changeFRAME,'down')     )
        self.vrt_frm_sldr.Bind(wx.EVT_SLIDER,    partial(self.changeFRAME,'vslider')   )

        aszr = wx.BoxSizer(wx.HORIZONTAL)
        bszr = wx.BoxSizer(wx.VERTICAL)
        cszr = wx.BoxSizer(wx.VERTICAL)
        dszr = wx.BoxSizer(wx.HORIZONTAL)

        aszr.Add(self.hrz_frm_btn[0],  flag=wx.RIGHT|wx.CENTER,  border=18)
        aszr.Add(self.hrz_frm_btn[1],  flag=wx.LEFT|wx.CENTER,   border=18)

        bszr.Add(self.vrt_frm_btn[0],  flag=wx.BOTTOM|wx.CENTER,    border=8)
        bszr.Add(aszr,             flag=wx.CENTER,              border=8)
        bszr.Add(self.vrt_frm_btn[1],  flag=wx.TOP|wx.CENTER,       border=8)

        cszr.Add(bszr,             flag=wx.CENTER,    border=6)
        cszr.Add(self.hrz_frm_sldr,  flag=wx.CENTER,    border=6)

        dszr.AddSpacer(50)
        dszr.Add(cszr,             flag=wx.CENTER,    border=6)
        dszr.Add(self.vrt_frm_sldr,  flag=wx.CENTER,    border=6)



        vbox.Add(dszr, flag=wx.EXPAND|wx.CENTER|wx.ALL, border=8)

        return vbox


    def Toolbox(self,panel):
        '''
        Frame for visual toolbox
        '''

        tlbx = wx.StaticBox(self.panel,label='2DXRD TOOLBOX')#, size=(200, 200))
        vbox = wx.StaticBoxSizer(tlbx,wx.VERTICAL)

        ###########################
        ## Color
        hbox_clr = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_clr = wx.StaticText(self.panel, label='COLOR')

        colors = []
        for key in colormap.datad:
            if not key.endswith('_r'):
                colors.append(key)
        self.ch_clr = wx.Choice(self.panel,choices=colors)
        #self.ch_clr = wx.Choice(self.panel,choices=ColorMap_List)

        self.ch_clr.Bind(wx.EVT_CHOICE,self.setCOLOR)

        hbox_clr.Add(self.txt_clr, flag=wx.RIGHT,  border=6)
        hbox_clr.Add(self.ch_clr,  flag=wx.RIGHT,  border=6)
        vbox.Add(hbox_clr,         flag=wx.ALL,    border=4)

        ###########################
        ## Contrast
        vbox_ct = wx.BoxSizer(wx.VERTICAL)

        hbox_ct1 = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_ct1 = wx.StaticText(self.panel, label='CONTRAST')
        self.txt_ct2 = wx.StaticText(self.panel, label='',   size=(175,-1))

        hbox_ct1.Add(self.txt_ct1, flag=wx.RIGHT,         border=6)
        hbox_ct1.Add(self.txt_ct2, flag=wx.RIGHT,         border=6)
        vbox_ct.Add(hbox_ct1,      flag=wx.TOP|wx.BOTTOM, border=4)

        hbox_ct2 = wx.BoxSizer(wx.HORIZONTAL)

        self.sldr_cntrst = wx.Slider(self.panel, style=wx.SL_LABELS,
                                     size=(275,-1), maxValue=5e6)

        self.entr_min = wx.TextCtrl(self.panel,  style=wx.TE_PROCESS_ENTER, size=(50,-1))
        self.entr_max = wx.TextCtrl(self.panel,  style=wx.TE_PROCESS_ENTER, size=(80,-1))

        self.sldr_cntrst.Bind(wx.EVT_SLIDER,self.onSlider)
        self.entr_min.Bind(wx.EVT_TEXT_ENTER,self.onContrastRange)
        self.entr_max.Bind(wx.EVT_TEXT_ENTER,self.onContrastRange)

        self.btn_ct1 = wx.Button(self.panel,label='reset',size=(50,-1))

        self.btn_ct1.Bind(wx.EVT_BUTTON,partial(self.setContrast,auto_contrast=True) )

        ttl_to = wx.StaticText(self.panel, label='to')
        hbox_ct2.Add(self.entr_min, flag=wx.RIGHT|wx.ALIGN_RIGHT, border=6)
        hbox_ct2.Add(ttl_to, flag=wx.RIGHT|wx.ALIGN_RIGHT, border=6)
        hbox_ct2.Add(self.entr_max, flag=wx.RIGHT|wx.ALIGN_RIGHT, border=6)
        hbox_ct2.Add(self.btn_ct1, flag=wx.RIGHT,              border=6)

        vbox_ct.Add(self.sldr_cntrst, flag=wx.EXPAND|wx.RIGHT,    border=6)
        vbox_ct.Add(hbox_ct2,         flag=wx.CENTER|wx.TOP,      border=6)
        vbox.Add(vbox_ct,             flag=wx.ALL,                border=4)

        ###########################
        ## Flip
        hbox_flp = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_flp = wx.StaticText(self.panel, label='IMAGE FLIP')
        flips = ['none','vertical','horizontal','both']
        self.ch_flp = wx.Choice(self.panel,choices=flips)

        self.ch_flp.Bind(wx.EVT_CHOICE,self.setFLIP)

        hbox_flp.Add(self.txt_flp, flag=wx.RIGHT|wx.TOP|wx.BOTTOM, border=6)
        hbox_flp.Add(self.ch_flp,  flag=wx.RIGHT|wx.TOP|wx.BOTTOM, border=6)
        vbox.Add(hbox_flp,         flag=wx.ALL,   border=4)

        ###########################
        ## Scale
        hbox_scl = wx.BoxSizer(wx.HORIZONTAL)
        self.txt_scl = wx.StaticText(self.panel, label='SCALE')
        scales = ['linear','log']
        self.ch_scl = wx.Choice(self.panel,choices=scales)

        self.ch_scl.Bind(wx.EVT_CHOICE,self.setZSCALE)

        hbox_scl.Add(self.txt_scl, flag=wx.RIGHT|wx.TOP|wx.BOTTOM, border=6)
        hbox_scl.Add(self.ch_scl,  flag=wx.RIGHT|wx.TOP|wx.BOTTOM, border=6)
        vbox.Add(hbox_scl,         flag=wx.ALL,   border=4)


        ###########################
        ## Mask
        hbox_msk = wx.BoxSizer(wx.HORIZONTAL)
        self.btn_mask = wx.Button(panel,label='MASK')
        self.ch_msk = wx.CheckBox(self.panel,label='Apply?')

        self.ch_msk.Bind(wx.EVT_CHECKBOX,self.applyMask)
        self.btn_mask.Bind(wx.EVT_BUTTON,self.openMask)

        hbox_msk.Add(self.btn_mask, flag=wx.RIGHT|wx.TOP|wx.BOTTOM, border=6)
        hbox_msk.Add(self.ch_msk,   flag=wx.RIGHT|wx.TOP|wx.BOTTOM, border=6)
        vbox.Add(hbox_msk,          flag=wx.ALL,   border=4)

        ###########################
        ## Background
        hbox_bkgd1 = wx.BoxSizer(wx.HORIZONTAL)
        self.btn_bkgd = wx.Button(panel,label='BACKGROUND')
        self.sldr_bkgd = wx.Slider(self.panel,style=wx.SL_VALUE_LABEL, maxValue=1e3)

        self.sldr_bkgd.Bind(wx.EVT_SLIDER,self.onBkgdScale)
        self.btn_bkgd.Bind(wx.EVT_BUTTON,self.openBkgd)

        hbox_bkgd1.Add(self.btn_bkgd,  flag=wx.RIGHT|wx.TOP,                 border=6)
        hbox_bkgd1.Add(self.sldr_bkgd, flag=wx.EXPAND|wx.ALIGN_RIGHT|wx.TOP, border=6)
        vbox.Add(hbox_bkgd1,           flag=wx.TOP|wx.BOTTOM,                border=4)

        self.sldr_bkgd.SetValue(self.bkgd_scale*SLIDER_SCALE)

        ###########################
        ## Set defaults
        self.ch_clr.SetStringSelection(self.color)
        self.ch_flp.SetStringSelection(self.flip)
        self.ch_scl.SetSelection(0)
        self.ch_msk.Disable()
        self.ch_clr.Disable()
        self.sldr_cntrst.Disable()
        self.entr_min.Disable()
        self.entr_max.Disable()
        self.btn_ct1.Disable()
        self.ch_flp.Disable()
        self.ch_scl.Disable()
        self.btn_mask.Disable()
        self.btn_bkgd.Disable()
        self.sldr_bkgd.Disable()

        self.hrz_frm_sldr.Disable()
        for btn in self.hrz_frm_btn: btn.Disable()

        self.vrt_frm_sldr.Disable()
        for btn in self.vrt_frm_btn: btn.Disable()

        return vbox


    def RightSidePanel(self,panel):
        vbox = wx.BoxSizer(wx.VERTICAL)

        self.xrd2Dcake   = diFFit2DPanel(panel,owner=self,size=(400,400),type='cake')
        self.xrd1Dviewer = diFFit1DPanel(panel,owner=self,size=(400,100))

        vbox.Add(self.xrd2Dcake,   proportion=1, flag=wx.TOP|wx.EXPAND,    border = 10)
        vbox.Add(self.xrd1Dviewer, proportion=1, flag=wx.BOTTOM|wx.EXPAND, border = 10)

        return vbox

    def CenterPanel(self,panel):

        self.xrd2Dviewer = diFFit2DPanel(panel,owner=self)
        btnbox = self.QuickButtons(panel)

        vbox = wx.BoxSizer(wx.VERTICAL)
        vbox.Add(self.xrd2Dviewer,proportion=1,flag=wx.ALL|wx.EXPAND,border = 10)
        vbox.Add(btnbox,flag=wx.ALL|wx.ALIGN_RIGHT,border = 10)
        return vbox

    def QuickButtons(self,panel):
        buttonbox = wx.BoxSizer(wx.HORIZONTAL)
        self.btn_img = wx.Button(panel,label='LOAD IMAGE')
        self.btn_calib = wx.Button(panel,label='CALIBRATION')
        self.btn_integ = wx.Button(panel,label='INTEGRATE (1D)')

        self.btn_img.Bind(wx.EVT_BUTTON,self.loadIMAGE)
        self.btn_calib.Bind(wx.EVT_BUTTON,self.openPONI)
        self.btn_integ.Bind(wx.EVT_BUTTON,self.on1DXRD)

        buttonbox.Add(self.btn_img, flag=wx.ALL, border=8)
        buttonbox.Add(self.btn_calib, flag=wx.ALL, border=8)
        buttonbox.Add(self.btn_integ, flag=wx.ALL, border=8)

        return buttonbox


class XRDImg(Group):
    '''
    XRD image class

    Attributes:
    ------------
    * self.label         = 'Data: CeO2_Allende'              # name of data set
    * self.path          = '/Volumes/Data/CeO2_Allende.tif'  # file containing x-y data
    * self.type          = 'tiff'                            # file type

    # Data parameters
    * self.image         = None or array        # should be a 3-D array [no * x * y]
    * self.iframes       = 1                    # number of frames in self.image   (row)
    * self.i             = 0                    # integer indicating current frame (row)
    * self.jframes       = 1                    # number of frames in self.image   (col)
    * self.j             = 0                    # integer indicating current frame (col)
    * self.minval        = 0                    # integer of minimum display contrast
    * self.maxval        = 100                  # integer of maximum display contrast

    mkak 2017.08.15
    '''

    def __init__(self, label=None, path='', type='tiff', image=None, h5file=None):

        self.label = label
        self.path  = path
        self.type  = type

        self.calfile = None

        self.h5file = h5file
        self.image = np.zeros((1,1,PIXELS,PIXELS)) if image is None else image

        self.check_image()
        self.calc_range()


    def check_image(self):

        if self.h5file is None:
            shp = np.shape(self.image)
            if len(shp) == 2:
                self.image = np.reshape(self.image,(1,1,shp[0],shp[1]))
            if len(shp) == 3:
                self.image = np.reshape(self.image,(1,shp[0],shp[1],shp[2]))

            self.jframes,self.iframes,self.xpix,self.ypix = np.shape(self.image)
        else:
            try:
                self.h5xrd = self.h5file['xrmmap/xrd2D/counts']
            except:
                self.image = None
                print('No 2DXRD data in %s' % os.path.split(self.h5file.filename)[-1])
                return
            self.calfile = bytes2str(self.h5file['xrmmap/xrd1D'].attrs.get('calfile',''))
            if not os.path.exists(self.calfile):
                self.calfile = None

            ## making an assumption that h5 map file always has multiple rows and cols
            self.jframes,self.iframes,self.xpix,self.ypix = self.h5xrd.shape

        self.i = 0 if self.iframes < 4 else int(self.iframes)/2
        self.j = 0 if self.jframes < 4 else int(self.jframes)/2

    def calc_range(self):

        if self.h5file is None:
            self.minval = self.image[self.j,self.i].min()
            self.maxval = self.image[self.j,self.i].max()
        else:
            self.minval = self.h5xrd[self.j,self.i].min()
            self.maxval = self.h5xrd[self.j,self.i].max()

    def get_image(self,i=None,j=None):

        if i is not None and i != self.i:
            if i < 0: i == self.iframes-1
            if i >= self.iframes: i = 0
            self.i = i

        if j is not None and j != self.j:
            if j < 0: j == self.jframes-1
            if j >= self.jframes: j = 0
            self.j = j

        if self.h5file is None:
            return self.image[self.j,self.i]
        else:
            return self.h5xrd[self.j,self.i]

    def set_contrast(self,minval,maxval):

        if maxval == minval: maxval = minval+100

        self.minval = int(minval)
        self.maxval = int(maxval)


class XRD2DViewer(wx.App):
    def __init__(self):
        wx.App.__init__(self)

    def createApp(self):
        frame = XRD2DViewerFrame()
        frame.Show()
        self.SetTopWindow(frame)

    def OnInit(self):
        self.createApp()
        return True



if __name__ == '__main__':
    XRD2DViewer().MainLoop()



#!/usr/bin/env python
import time
from datetime import datetime
from collections import OrderedDict
import numpy as np
import copy
import json

from .paths import uname, bindir, nativepath, get_homedir
from .debugtime import debugtime, debugtimer

from .strutils import (fixName, isValidName, isNumber, bytes2str,
                       fix_varname, isLiteralStr, strip_comments,
                       find_delims, version_ge, unique_name)

from .shellutils import (_copy, _deepcopy, _more, _parent,
                         _ls, _cd, _cwd, _mkdir)


def group2dict(group, _larch=None):
    "return dictionary of group members"
    return group.__dict__

def dict2group(d, _larch=None):
    "return group created from a dictionary"
    from larch import Group
    return Group(**d)

def copy_group(group, _larch=None):
    from larch import Group
    out = Group(datatype=getattr(group, 'datatype', 'unknown'),
                copied_from=getattr(group, 'groupname', repr(group)))

    class NoCopy:
        c = 'no copy'
        
    for attr in dir(group):
        val = NoCopy
        try:
            val = copy.deepcopy(getattr(group, attr))
        except ValueError:
            try:
                val = copy.copy(getattr(group, attr))
            except:
                val = NoCopy
        
        if val != NoCopy:
            setattr(out, attr, val)
    return out

def copy_xafs_group(group, _larch=None):
    """specialized group copy for XAFS data groups"""
    from larch import Group
    out = Group(datatype=getattr(group, 'datatype', 'unknown'),
                copied_from=getattr(group, 'groupname', repr(group)))

    for attr in dir(group):
        do_copy = True
        if attr in ('xdat', 'ydat', 'i0', 'data' 'yerr',
                    'energy', 'mu'):
            val = getattr(group, attr)*1.0
        elif attr in ('norm', 'flat', 'deriv', 'deconv',
                      'post_edge', 'pre_edge', 'norm_mback',
                      'norm_vict', 'norm_poly'):
            do_copy = False
        else:
            try:
                val = copy.deepcopy(getattr(group, attr))
            except ValueError:
                do_copy = False
        if do_copy:
            setattr(out, attr, val)
    return out
   

def isotime(t=None, with_tzone=False):
    if t is None:
        t = time.time()
    sout = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(t))
    if with_tzone:
        sout = "%s-%2.2i:00" % (sout, time.timezone/3600)
    return sout

def json_dump(data, filename):
    """
    dump object or group to file using json
    """
    from .jsonutils import encode4js
    with open(filename, 'w') as fh:
        fh.write(json.dumps(encode4js(data)))
        fh.write('\n')

def json_load(filename):
    """
    load object from json dump file
    """
    from .jsonutils import decode4js
    with open(filename, 'r') as fh:
        data = fh.read()
    return decode4js(json.loads(data))

def _larch_init(_larch):
    """initialize xrf"""
    from ..symboltable import Group
    _larch.symtable._sys.display = Group(use_color=True,
                                         colors=dict(text={'color': 'black'},
                                                     text2={'color': 'blue'},
                                                     error={'color': 'red'}))

_larch_builtins = dict(copy=_copy, deepcopy=_deepcopy, more= _more,
                       parent=_parent, ls=_ls, mkdir=_mkdir, cd=_cd,
                       cwd=_cwd, group2dict=group2dict,
                       copy_group=copy_group, copy_xafs_group=copy_xafs_group,
                       dict2group=dict2group, debugtimer=debugtimer,
                       isotime=isotime, json_dump=json_dump,
                       json_load=json_load)



#!/usr/bin/env python
"""
Spyk is a Spec emulator, providing Spec-like scanning functions to Larch

    spec = SpykScan()
    spec.add_motors(x='XX:m1', y='XX:m2')
    spec.add_detector('XX:scaler1')
    spec.set_scanfile(outputfile)

    spec.ascan('x', start, finish, npts, time)
    spec.a2scan('x', s1, f1, 'y', s1, f1, npts, time)
    spec.a3scan('x', s1, f1, 'y', s2, f2, 'z', s3, f3, npts, time)
    spec.mesh('x', s1, f1, npts1, 'y', s2, f2, npts2, time)

    spec.lup('x', start, finish, npts, time)
    spec.dscan('x', start, finish, npts, time)
    spec.d2scan('x', s1, f1, 'y', s1, f1, npts, time)
    spec.d3scan('x', s1, f1, 'y', s2, f2, 'z', s3, f3, npts, time)

yet to be implemented:
    -- th2th tth_start_rel tth_finish_rel intervals time
    -- automatic plotting
    -- save/read configuration
"""
import os
import sys
import time
import numpy as np
from collections import OrderedDict
from io import StringIO
from configparser import ConfigParser


from larch import Group

from larch.utils import get_homedir
from larch.io import get_timestamp

try:
    from epics import PV, caget, caput, get_pv, poll
    HAS_EPICS = True
except ImportError:
    HAS_EPICS = False

try:
    from .larchscan   import LarchStepScan
    import epicsscan
    from epicsscan.positioner import Positioner
    from epicsscan.detectors  import get_detector, Counter
    HAS_EPICSSCAN = True
except ImportError:
    HAS_EPICSSCAN = False


class SpykConfig(object):
    """
    Configuration file (INI format) for Spyk scanning
    """
    LEGEND     = '# index = label || PVname'
    DET_LEGEND = '# index = label || DetectorPV || options '
    SPYK_DIR   = '.spyk'
    SPYK_INI   = 'spyk.ini'
    #  sections            name      ordered?
    __sects = OrderedDict((('setup',     False),
                           ('motors',    True),
                           ('detectors', True),
                           ('extra_pvs', True),
                           ('counters',  True)))

    def __init__(self, filename=None, text=None):
        for s in self.__sects:
            setattr(self, s, {})
        self._cp = ConfigParser()
        self.filename = filename
        if self.filename is None:
            cfile = self.get_default_configfile()
            if (os.path.exists(cfile) and os.path.isfile(cfile)):
                self.filename = cfile
        if self.filename is not None:
            self.Read(self.filename)

    def get_default_configfile(self):
        return os.path.join(get_homedir(), self.SPYK_DIR, self.SPYK_INI)

    def Read(self, fname=None):
        "read config"
        if fname is None:
            return
        ret = self._cp.read(fname)
        if len(ret) == 0:
            time.sleep(0.25)
            ret = self._cp.read(fname)
        self.filename = fname
        # process sections
        for sect, ordered in self.__sects.items():
            if not self._cp.has_section(sect):
                continue
            thissect = OrderedDict() if ordered else {}
            for opt in self._cp.options(sect):
                val = self._cp.get(sect, opt)
                if '||' in val:
                    words = [i.strip() for i in val.split('||')]
                    label = words.pop(0)
                    if len(words) == 1:
                        words = words[0]
                    else:
                        words = tuple(words)
                    thissect[label] = words
                else:
                    thissect[opt] = val
                setattr(self, sect, thissect)

    def Save(self, fname=None):
        "save config file"
        if fname is None:
            fname = self.get_default_configfile()
            path, fn = os.path.split(fname)
            if not os.path.exists(path):
                os.makedirs(path, mode=755)

        out = ['###Spyke Configuration: %s'  % (get_timestamp())]
        for sect, ordered in self.__sects.items():
            out.append('#-----------------------#\n[%s]' % sect)
            if sect == 'setup':
                for name, val in self.setup.items():
                    out.append("%s = %s" % (name, val))
            elif sect == 'detectors':
                out.append(self.DET_LEGEND)
                idx = 0
                for key, val in getattr(self, sect).items():
                    idx = idx + 1
                    if isinstance(val, (list, tuple)):
                        val = ' || '.join(val)
                    out.append("%i = %s || %s"  % (idx, key, val))

            else:
                out.append(self.LEGEND)
                idx = 0
                for key, val in getattr(self, sect).items():
                    idx = idx + 1
                    if isinstance(val, (list, tuple)):
                        val = ' || '.join(val)
                    out.append("%i = %s || %s"  % (idx, key, val))
        out.append('#-----------------------#')
        with open(fname, 'w') as fh:
            fh.write('\n'.join(out))

    def sections(self):
        return self.__sects.keys()


class Spyk(Group):
    """Spyk is a set of Spec-like scanning tools for Larch"""
    def __init__(self, filename='spykscan.001', configfile=None,
                 auto_increment=True, _larch=None, **kwargs):
        Group.__init__(self, **kwargs)
        self.motors  = {}
        self.detectors = []
        self.bare_counters = []
        self._scan = LarchStepScan(filename=filename,
                                   auto_increment=auto_increment,
                                   _larch=_larch)
        self.datafilename = filename
        if configfile is not None:
            self.configfile = configfile
        self.read_config(filename=configfile)
        self.lup = self.dscan

    def read_config(self, filename=None):
        "read Spyk configuration file"
        self.config = SpykConfig(filename=filename)
        self.configfile = self.config.filename
        for label, pvname in self.config.motors.items():
            self.motors[label] = Positioner(pvname, label=label)

        for label, pvname in self.config.counters.items():
            self.bare_counters.append(Counter(pvname, label=label))

        self.add_extra_pvs(self.config.extra_pvs.items())
        for label, val in self.config.detectors.items():
            opts = {'label': label}
            prefix = val[0]
            if len(val) > 1:
                pairs = [w.strip() for w in val[1].split(',')]
                for s in pairs:
                    skey, sval = s.split('=')
                    opts[skey.strip()] = sval.strip()
            self.add_detector(prefix, **opts)

    def save_config(self, filename=None):
        "save Spyk configuration file"
        print( 'save spyk config....')

    def add_motors(self, **motors):
        """add motors as keyword=value pairs: label=EpicsPVName"""
        for label, pvname in motors.items():
            self.motors[label] = Positioner(pvname, label=label)

    def add_counter(self, name, label=None):
        # self._scan.add_counter(name, label=label)
        self.bare_counters.append(Counter(name, label=label))

    def add_detector(self, name, kind=None, **kws):
        "add detector, giving base name and detector type"
        self.detectors.append(get_detector(name, kind=kind, **kws))

    def add_extra_pvs(self, extra_pvs):
        """add extra PVs to be recorded prior to each scan
        extra_pvs should be list or tuple of (label, PVname)
        """
        self._scan.add_extra_pvs(extra_pvs)

    def set_scanfile(self, filename):
        "set file name"
        self.datafilename = filename

    def _checkmotors(self, *args):
        "check that all args are motor names"
        for mname in args:
            if mname not in self.motors:
                raise Exception("Error: unknown motor name '%s'" % mname)

    def _run(self, dwelltime):
        """internal function to start scans"""
        self._scan.counters = []
        self._scan.triggers = []
        self._scan.dwelltime = dwelltime
        for d in self.detectors:
            self._scan.add_detector(d)
            d.dwelltime = dwelltime
        for c in self.bare_counters:
            self._scan.add_counter(c)

        self._scan.run(filename=self.datafilename)

    def ascan(self, motor, start, finish, npts, dtime):
        "ascan: absolute scan"
        self._checkmotors(motor)
        self._scan.positioners  = [self.motors[motor]]
        self._scan.positioners[0].array = np.linspace(start, finish, npts)
        self._run(dtime)

    def dscan(self, motor, start, finish, npts, dtime):
        "dscan: relative scan"
        self._checkmotors(motor)
        current = self.motors[motor].current()
        start  += current
        finish += current
        self.ascan(motor, start, finish, npts, dtime)

    def a2scan(self, motor1, start1, finish1,
               motor2, start2, finish2, npts, dtime):
        "a2scan: absolute scan of 2 motors"
        self._checkmotors(motor1, motor2)
        self._scan.positioners  = [self.motors[motor1], self.motors[motor2]]
        self._scan.positioners[0].array = np.linspace(start1, finish1, npts)
        self._scan.positioners[1].array = np.linspace(start2, finish2, npts)
        self._run(dtime)

    def d2scan(self, motor1, start1, finish1,
               motor2, start2, finish2, npts, dtime):
        "d2scan: relative scan of 2 motors"
        self._checkmotors(motor1, motor2)
        current1 = self.motors[motor1].current()
        start1  += current1
        finish1 += current1

        current2 = self.motors[motor2].current()
        start2  += current2
        finish2 += current2

        self.a2scan(motor1, start1, finish1,
                    motor2, start2, finish2, npts, dtime)

    def a3scan(self, motor1, start1, finish1, motor2, start2, finish2,
               motor3, start3, finish3, npts, dtime):
        "a3scan: absolute scan of 3 motors"
        self._checkmotors(motor1, motor2, motor3)

        self._scan.positioners  = [self.motors[motor1],
                                   self.motors[motor2],
                                   self.motors[motor3]]
        self._scan.positioners[0].array = np.linspace(start1, finish1, npts)
        self._scan.positioners[1].array = np.linspace(start2, finish2, npts)
        self._scan.positioners[2].array = np.linspace(start3, finish3, npts)
        self._run(dtime)

    def d3scan(self, motor1, start1, finish1, motor2, start2, finish2,
               motor3, start3, finish3, npts, dtime):
        "d3scan: relative scan of 3 motors"
        self._checkmotors(motor1, motor2, motor3)

        current1 = self.motors[motor1].current()
        start1  += current1
        finish1 += current1

        current2 = self.motors[motor2].current()
        start2  += current2
        finish2 += current2

        current3 = self.motors[motor2].current()
        start3  += current3
        finish3 += current3

        self.a3scan(motor1, start1, finish1,
                    motor2, start2, finish2,
                    motor3, start3, finish3, npts, dtime)

    def mesh(self, motor1, start1, finish1, npts1,
               motor2, start2, finish2, npts2, dtime):
        """mesh scan: absolute scan of motor1 at each
        position for motor2"""
        self._checkmotors(motor1, motor2)

        self._scan.positioners = [self.motors[motor1], self.motors[motor2]]

        fast = npts2* [np.linspace(start1, finish1, npts1)]
        slow = [[i]*npts1 for i in np.linspace(start2, finish2, npts2)]

        self._scan.positioners[0].array = np.array(fast).flatten()
        self._scan.positioners[1].array = np.array(slow).flatten()

        # set breakpoints to be the end of each row
        self._scan.breakpoints = [(i+1)*npts1 - 1 for i in range(npts2-1)]

        # add print statement at end of each row
        def show_meshstatus(breakpoint=None):
            print('finished row  %i of %i' % (1+(breakpoint/npts1), npts2))
            time.sleep(0.25)
        self._scan.at_break_methods.append(show_meshstatus)
        self._run(dtime)
